from math import ceil
from posixpath import join
import numpy as np
from tomoscan.io import HDF5File
from .cli_configs import CompareVolumesConfig
from ..utils import clip_circle
from .utils import parse_params_values
from ..io.utils import get_first_hdf5_entry, hdf5_entry_exists
from ..io.reader import get_hdf5_dataset_shape


def idx_1d_to_3d(idx, shape):
    nz, ny, nx = shape
    x = idx % nx
    idx2 = (idx - x) // nx
    y = idx2 % ny
    z = (idx2 - y) // ny
    return (z, y, x)


def compare_volumes(fname1, fname2, h5_path, chunk_size, do_stats, stop_at_thresh, clip_outer_circle=False):
    result = None
    f1 = HDF5File(fname1, "r")
    f2 = HDF5File(fname2, "r")
    try:
        # Check that data is in the provided hdf5 path
        for fname in [fname1, fname2]:
            if not hdf5_entry_exists(fname, h5_path):
                result = "File %s do not has data in %s" % (fname, h5_path)
                return
        # Check shapes
        shp1 = get_hdf5_dataset_shape(fname1, h5_path)
        shp2 = get_hdf5_dataset_shape(fname2, h5_path)
        if shp1 != shp2:
            result = "Volumes do not have the same shape: %s vs %s" % (shp1, shp2)
            return
        # Compare volumes
        n_steps = ceil(shp1[0] / chunk_size)
        for i in range(n_steps):
            start = i * chunk_size
            end = min((i + 1) * chunk_size, shp1[0])
            data1 = f1[h5_path][start:end, :, :]
            data2 = f2[h5_path][start:end, :, :]
            abs_diff = np.abs(data1 - data2)
            if clip_outer_circle:
                for j in range(abs_diff.shape[0]):
                    abs_diff[j] = clip_circle(abs_diff[j], radius=0.9 * min(abs_diff.shape[1:]))
            coord_argmax = idx_1d_to_3d(np.argmax(abs_diff), data1.shape)
            if do_stats:
                mean = np.mean(abs_diff)
                std = np.std(abs_diff)
                maxabs = np.max(abs_diff)
                print("Chunk %d: mean = %e std = %e max = %e" % (i, mean, std, maxabs))
            if stop_at_thresh is not None and abs_diff[coord_argmax] > stop_at_thresh:
                coord_argmax_absolute = (start + coord_argmax[0],) + coord_argmax[1:]
                result = "abs_diff[%s] = %e" % (coord_argmax_absolute, abs_diff[coord_argmax])
                return
    except Exception as exc:
        result = "Error: %s" % (str(exc))
        raise
    finally:
        f1.close()
        f2.close()
        return result


def compare_volumes_cli():
    args = parse_params_values(
        CompareVolumesConfig, parser_description="A command-line utility for comparing two volumes."
    )

    fname1 = args["volume1"]
    fname2 = args["volume2"]
    h5_path = args["hdf5_path"]
    if h5_path == "":
        entry = args["entry"].strip() or None
        if entry is None:
            entry = get_first_hdf5_entry(fname1)
        h5_path = join(entry, "reconstruction/results/data")
    do_stats = bool(args["statistics"])
    chunk_size = args["chunk_size"]
    stop_at_thresh = args["stop_at"] or None
    if stop_at_thresh is not None:
        stop_at_thresh = float(stop_at_thresh)

    res = compare_volumes(fname1, fname2, h5_path, chunk_size, do_stats, stop_at_thresh)
    if res is not None:
        print(res)


if __name__ == "__main__":
    compare_volumes_cli()



from math import log2, ceil
import numpy as np
from silx.math import Histogramnd
from tomoscan.io import HDF5File
from ..utils import check_supported
from ..resources.logger import LoggerOrPrint


class PartialHistogram:
    """
    A class for computing histogram progressively.

    In certain cases, it is cumbersome to compute a histogram directly on a big chunk of
    data (ex. data not fitting in memory, disk access too slow) while some parts of the
    data are readily available in-memory.
    """

    histogram_methods = ["fixed_bins_width", "fixed_bins_number"]
    bin_width_policies = ["uint16"]
    backends = ["numpy", "silx"]

    def __init__(self, method="fixed_bins_width", bin_width="uint16", num_bins=None, min_bins=None, backend="silx"):
        """
        Initialize a PartialHistogram class.

        Parameters
        ----------
        method: str, optional
            Partial histogram computing method. Available are:
               - `fixed_bins_width`: all the histograms are computed with the same bin
                 width. The class adapts to the data range and computes the number of
                 bins accordingly.
               - `fixed_bins_number`: all the histograms are computed with the same
                 number of bins. The class adapts to the data range and computes the
                 bin width accordingly.
            Default is "fixed_bins_width"
        bin_width: str or float, optional
            Policy for histogram bins when method="fixed_bins_width". Available are:
               - "uint16": The bin width is computed so that floating-point elements
                 `f1` and `f2` satisfying `|f1 - f2| < bin_width` implies
                 `f1_converted - f2_converted < 1` once cast to uint16.
               - A number: all the bins have this fixed width.

            Default is "uint16"
        num_bins: int, optional
            Number of bins when method = 'fixed_bins_number'.
        min_bins: int, optional
            Minimum number of bins when method = 'fixed_bins_width'.
        backend: str, optional
            Which histogram backend to use for computations. Available are "silx", "numpy".
            Fastest is "silx".
        """
        check_supported(method, self.histogram_methods, "histogram computing method")
        self.method = method
        check_supported(backend, self.backends, "histogram backend")
        self.backend = backend
        self._set_bin_width(bin_width)
        self._set_num_bins(num_bins)
        self.min_bins = min_bins
        self._set_histogram_methods()

    def _set_bin_width(self, bin_width):
        if self.method == "fixed_bins_number":
            self.bin_width = None
            return
        if isinstance(bin_width, str):
            check_supported(bin_width, self.bin_width_policies, "bin width policy")
            self._fixed_bw = False
        else:
            bin_width = float(bin_width)
            self._fixed_bw = True
        self.bin_width = bin_width

    def _set_num_bins(self, num_bins):
        if self.method == "fixed_bins_width":
            self.num_bins = None
            return
        if self.method == "fixed_bins_number" and num_bins is None:
            raise ValueError("Need to specify num_bins for method='fixed_bins_number'")
        self.num_bins = int(num_bins)

    def _set_histogram_methods(self):
        self._histogram_methods = {
            "fixed_bins_number": {
                "compute": self._compute_histogram_fixed_nbins,
                "merge": self._merge_histograms_fixed_nbins,
            },
            "fixed_bins_width": {
                "compute": self._compute_histogram_fixed_bw,
                "merge": self._merge_histograms_fixed_bw,
            },
        }
        assert set(self._histogram_methods.keys()) == set(self.histogram_methods)

    @staticmethod
    def _get_histograms_and_bins(histograms, center=False, dont_truncate_bins=False):
        histos = [h[0] for h in histograms]
        if dont_truncate_bins:
            bins = [h[1] for h in histograms]
        else:
            if center:
                bins = [0.5 * (h[1][1:] + h[1][:-1]) for h in histograms]
            else:
                bins = [h[1][:-1] for h in histograms]
        return histos, bins

    #
    # Histogram with fixed number of bins
    #

    def _compute_histogram_fixed_nbins(self, data, data_range=None):
        if data.ndim > 1:
            data = data.ravel()
        dmin, dmax = data.min(), data.max() if data_range is None else data_range
        if self.backend == "numpy":
            res = np.histogram(data, bins=self.num_bins)
        elif self.backend == "silx":
            histogrammer = Histogramnd(data, n_bins=self.num_bins, histo_range=(dmin, dmax), last_bin_closed=True)
            res = histogrammer.histo, histogrammer.edges[0]  # pylint: disable=E1136
        return res

    def _merge_histograms_fixed_nbins(self, histograms, dont_truncate_bins=False):
        histos, bins = self._get_histograms_and_bins(histograms, dont_truncate_bins=dont_truncate_bins)
        res = np.histogram(
            np.hstack(bins),
            weights=np.hstack(histos),
            bins=self.num_bins,
        )
        return res

    #
    # Histogram with fixed bin width
    #

    def _bin_width_u16(self, dmin, dmax):
        return (dmax - dmin) / 65535.0

    def _bin_width_fixed(self, dmin, dmax):
        return self.bin_width

    def get_bin_width(self, dmin, dmax):
        if self._fixed_bw:
            return self._bin_width_fixed(dmin, dmax)
        elif self.bin_width == "uint16":
            return self._bin_width_u16(dmin, dmax)
        else:
            raise ValueError()

    def _compute_histogram_fixed_bw(self, data, data_range=None):
        dmin, dmax = data.min(), data.max() if data_range is None else data_range
        min_bins = self.min_bins or 1
        bw_max = self.get_bin_width(dmin, dmax)
        nbins = 0
        bw_factor = 1
        while nbins < min_bins:
            bw = 2 ** round(log2(bw_max)) / bw_factor
            nbins = int((dmax - dmin) / bw)
            bw_factor *= 2
        res = np.histogram(data, bins=nbins)
        return res

    def _merge_histograms_fixed_bw(self, histograms, **kwargs):
        histos, bins = self._get_histograms_and_bins(histograms, center=False)
        dmax = max([b[-1] for b in bins])
        dmin = min([b[0] for b in bins])
        bw_max = max([b[1] - b[0] for b in bins])
        res = np.histogram(np.hstack(bins), weights=np.hstack(histos), bins=int((dmax - dmin) / bw_max))
        return res

    #
    # Dispatch methods
    #

    def compute_histogram(self, data, data_range=None):
        compute_hist_func = self._histogram_methods[self.method]["compute"]
        return compute_hist_func(data, data_range=data_range)

    def merge_histograms(self, histograms, **kwargs):
        merge_hist_func = self._histogram_methods[self.method]["merge"]
        return merge_hist_func(histograms, **kwargs)


class VolumeHistogram:
    """
    A class for computing the histogram of an entire volume.
    Unless explicitly specified, histogram is computed in several passes so that not
    all the volume is loaded in memory.
    """

    def __init__(self, data_url, chunk_size_slices=100, chunk_size_GB=None, nbins=1e6, logger=None):
        """
        Initialize a VolumeHistogram object.

        Parameters
        ----------
        fname: DataUrl
            DataUrl to the HDF5 file.
        chunk_size_slices: int, optional
            Compute partial histograms of groups of slices. This is the default behavior,
            where the groups size is 100 slices.
            This parameter is mutually exclusive with 'chunk_size_GB'.
        chunk_size_GB: float, optional
            Maximum memory (in GB) to use when computing the histogram by group of slices.
            This parameter is mutually exclusive with 'chunk_size_slices'.
        nbins: int, optional
            Histogram number of bins. Default is 1e6.
        """
        self.data_url = data_url
        self.logger = LoggerOrPrint(logger)
        self._get_data_info()
        self._set_chunk_size(chunk_size_slices, chunk_size_GB)
        self.nbins = int(nbins)
        self._init_histogrammer()

    def _get_data_info(self):
        self.fname = self.data_url.file_path()
        self.data_path = self.data_url.data_path()
        with HDF5File(self.fname, "r") as fid:
            try:
                data_ptr = fid[self.data_path]
            except KeyError:
                msg = str(
                    "Could not access HDF5 path %s in file %s. Please check that this file \
                    actually contains a reconstruction and that the HDF5 path is correct"
                    % (self.data_path, self.fname)
                )
                self.logger.fatal(msg)
                raise ValueError(msg)
            if data_ptr.ndim != 3:
                msg = "Expected data to have 3 dimensions, got %d" % data_ptr.ndim
                raise ValueError(msg)
            self.data_shape = data_ptr.shape
            self.data_dtype = data_ptr.dtype
            self.data_nbytes_GB = np.prod(data_ptr.shape) * data_ptr.dtype.itemsize / 1e9

    def _set_chunk_size(self, chunk_size_slices, chunk_size_GB):
        if not ((chunk_size_slices is not None) ^ (chunk_size_GB is not None)):
            raise ValueError("Please specify either chunk_size_slices or chunk_size_GB")
        if chunk_size_slices is None:
            chunk_size_slices = int(chunk_size_GB / (np.prod(self.data_shape[1:]) * self.data_dtype.itemsize / 1e9))
        self.chunk_size = chunk_size_slices
        self.logger.debug("Computing histograms by groups of %d slices" % self.chunk_size)

    def _init_histogrammer(self):
        self.histogrammer = PartialHistogram(method="fixed_bins_number", num_bins=self.nbins)

    def _compute_histogram(self, data):
        return self.histogrammer.compute_histogram(data.ravel())  # 1D

    def compute_volume_histogram(self):
        n_z = self.data_shape[0]
        histograms = []
        n_steps = ceil(n_z / self.chunk_size)
        with HDF5File(self.fname, "r") as fid:
            for chunk_id in range(n_steps):
                self.logger.debug("Computing histogram %d/%d" % (chunk_id + 1, n_steps))
                z_slice = slice(chunk_id * self.chunk_size, (chunk_id + 1) * self.chunk_size)
                images_stack = fid[self.data_path][z_slice, :, :]
                hist = self._compute_histogram(images_stack)
                histograms.append(hist)
        res = self.histogrammer.merge_histograms(histograms)
        return res


def hist_as_2Darray(hist, center=True, dtype="f"):
    hist, bins = hist
    if bins.size != hist.size:
        # assert bins.size == hist.size +1
        if center:
            bins = 0.5 * (bins[1:] + bins[:-1])
        else:
            bins = bins[:-1]
    res = np.zeros((2, hist.size), dtype=dtype)
    res[0] = hist
    res[1] = bins.astype(dtype)
    return res


def add_last_bin(histo_bins):
    """
    Add the last bin (max value) to a list of bin edges.
    """
    res = np.zeros(histo_bins.size + 1, dtype=histo_bins.dtype)
    res[:-1] = histo_bins[:]
    res[-1] = res[-2] + (res[1] - res[0])
    return res



from os import path
from pathlib import Path as pathlib_Path
from posixpath import join as posixjoin
from silx.io.dictdump import dicttonx
from tomoscan.esrf import HDF5Volume, TIFFVolume, MultiTIFFVolume, EDFVolume, JP2KVolume
from ..resources.logger import LoggerOrPrint
from ..io.writer import get_datetime, NXProcessWriter, HSTVolVolume
from ..io.utils import convert_dict_values
from .. import version as nabu_version
from ..resources.utils import is_hdf5_extension

#
# There are still multiple issues:
#   - When using HDF5, we still have to do self.file_prefix += str("_%05d" % self.start_index)
#     because we are writing partial files. This should be done by tomoscan but it's likely incompatible
#     with its current architecture
#   - _configure_metadata() is too long, and somehow duplicates what is done in LegacyNXProcessWriter.
#   - LegacyNXProcessWriter still has to be used for writing 1D data (histogram)
#
#  All in all, tomoscan.esrf.volume does not seem to make things simpler, at least for HDF5.
#


class WriterManager:

    """
    This class is a wrapper on top of all "writers".
    It will create the right "writer" with all the necessary options, and the histogram writer.

    The layout is the following.

    For "single-file volume" formats (HDF5, big tiff):
       bigtiff:
         output_dir/file_prefix.tiff (bigtiff)
       hdf5:
         [output_dir/file_prefix/file_prefix_{%05d}.h5]   (partial results)
         output_dir/file_prefix.h5 (master file)

    For "one file per slice" formats (tiff, jp2, edf)
      output_dir/file_prefix_%05d.{ext}


    When saving intermediate steps (eg. sinogram): HDF5 format is always used.
    So the layout is

      [output_dir/sinogram_file_prefix/sinogram_file_prefix_%05d.h5] (partial results)
      output_dir/sinogram_file_prefix.h5 (master file)
    """

    _overwrite_warned = False

    def __init__(
        self,
        output_dir,
        file_prefix,
        file_format="hdf5",
        overwrite=False,
        start_index=0,
        logger=None,
        metadata=None,
        histogram=False,
        extra_options=None,
    ):
        """
        Create a Writer from a set of parameters.

        Parameters
        ----------
        output_dir: str
            Directory where the file(s) will be written.
        file_prefix: str
            File prefix (without leading path)
        start_index: int, optional
            Index to start the files numbering (filename_0123.ext).
            Default is 0.
            Ignored for HDF5 extension.
        logger: nabu.resources.logger.Logger, optional
            Logger object
        metadata: dict, optional
            Metadata, eg. information on various processing steps. For HDF5, it will go to "configuration"
        histogram: bool, optional
            Whether to also write a histogram of data. If set to True, it will configure
            an additional "writer".
        extra_options: dict, optional
            Other advanced options to pass to Writer class.
        """
        self.overwrite = overwrite
        self.start_index = start_index
        self.logger = LoggerOrPrint(logger)
        self.histogram = histogram
        self.extra_options = extra_options or {}

        self.is_hdf5_output = is_hdf5_extension(file_format)
        self.is_bigtiff = file_format in ["tiff", "tif"] and any(
            [self.extra_options.get(opt, False) for opt in ["tiff_single_file", "use_bigtiff"]]
        )
        self.is_vol = file_format == "vol"

        self.file_prefix = file_prefix
        self._set_output_dir(output_dir)
        self._set_file_name(file_format)
        self._configure_metadata(metadata)

        # tomoscan.esrf.volume arguments
        def _get_writer_kwargs_single_frame():
            return {
                "folder": self.output_dir,
                "volume_basename": self.file_prefix,
                "start_index": self.start_index,
                "overwrite": self.overwrite,
            }

        def _get_writer_kwargs_multi_frames():
            return {
                "file_path": self.fname,
                "overwrite": self.overwrite,
            }

        if self.is_hdf5_output:
            writer = HDF5Volume
            writer_kwargs = _get_writer_kwargs_multi_frames()
            writer_kwargs.update(
                {
                    "data_path": posixjoin(self._h5_entry, self._h5_process_name),
                }
            )
        elif file_format in ["tiff", "tif"]:
            if self.is_bigtiff:
                writer = MultiTIFFVolume
                writer_kwargs = _get_writer_kwargs_multi_frames()
                writer_kwargs.update({"append": self.extra_options.get("single_output_file_initialized", False)})
            else:
                writer = TIFFVolume
                writer_kwargs = _get_writer_kwargs_single_frame()
                writer_kwargs.update(
                    {
                        "folder": self.output_dir,
                        "volume_basename": self.file_prefix,
                        "overwrite": True,
                    }
                )
        elif file_format == "vol":
            writer = HSTVolVolume
            writer_kwargs = _get_writer_kwargs_multi_frames()
            writer_kwargs.update({"append": self.extra_options.get("single_output_file_initialized", False)})
        elif file_format == "edf":
            writer = EDFVolume
            writer_kwargs = _get_writer_kwargs_single_frame()
        elif file_format in ["jp2k", "j2k", "jp2", "jp2000", "jpeg2000"]:
            writer = JP2KVolume
            writer_kwargs = _get_writer_kwargs_single_frame()
        else:
            raise ValueError("Unsupported file format: %s" % file_format)
        self.writer = writer(**writer_kwargs)
        self._init_histogram_writer()

    def _set_output_dir(self, output_dir):
        # This class is generally used to create partial files, i.e files containing a subset of the processed volume.
        # In this case, the files containing partial results are stored in a sub-directory with the same file prefix.
        # Otherwise, everything is put in a single file (for now it's only the case for "big tiff").
        self.is_partial_file = not (self.is_bigtiff or self.is_vol)
        if self.is_partial_file:
            output_dir = path.join(output_dir, self.file_prefix)

        self.output_dir = output_dir
        if path.exists(self.output_dir):
            if not (path.isdir(self.output_dir)):
                raise ValueError(
                    "Unable to create directory %s: already exists and is not a directory" % self.output_dir
                )
        else:
            self.logger.debug("Creating directory %s" % self.output_dir)
            pathlib_Path(self.output_dir).mkdir(parents=True, exist_ok=True)

    def _set_file_name(self, file_format):
        if self.is_hdf5_output:
            # HDF5Volume() does not have a "start_index" parameter, so we have to handle it here
            # (HDF5 files are partial files that are eventually merged into a master file,
            # so they have a _%05d suffix)
            self.file_prefix += str("_%05d" % self.start_index)
        self.file_format = file_format
        self.fname = path.join(self.output_dir, self.file_prefix + "." + file_format)
        if path.exists(self.fname):
            err = "File already exists: %s" % self.fname
            if self.overwrite:
                if not (self.__class__._overwrite_warned):
                    self.logger.warning(err + ". It will be overwritten as requested in configuration")
                    self.__class__._overwrite_warned = True
            else:
                self.logger.fatal(err)
                raise ValueError(err)

    def _configure_metadata(self, metadata):
        metadata = metadata or {}
        self.metadata = convert_dict_values(metadata, {None: "None"})
        self._h5_entry = self.metadata.pop("entry", "entry")
        if self.is_hdf5_output:
            self.metadata.update({"@NX_class": "NXcollection"})  # should be done by tomoscan...
            self._h5_process_name = process_name = self.metadata.pop("process_name", "reconstruction")
            # Metadata in {entry}/reconstruction. Can be written now.
            nabu_process_info = {
                "@NX_class": "NXentry",
                "@default": f"{process_name}/results",
                f"{process_name}@NX_class": "NXprocess",
                f"{process_name}/program": "nabu",
                f"{process_name}/version": nabu_version,
                f"{process_name}/date": get_datetime(),
                f"{process_name}/sequence_index": self.metadata.pop("processing_index", 0),
                f"{process_name}@default": "results",
            }
            dicttonx(
                nabu_process_info,
                h5file=self.fname,
                h5path=self._h5_entry,
                update_mode="replace",
                mode="a",
            )
            # Metadata in {entry}/reconstruction/results. Will be written after data.
            self._h5_results_metadata = {
                f"{process_name}/results@NX_class": "NXdata",
                f"{process_name}/results@signal": "data",  # TODO "data_path" ?
                f"{process_name}/results@interpretation": "image",
                f"{process_name}/results/data@interpretation": "image",
            }

    def _init_histogram_writer(self):
        if not self.histogram:
            return
        separate_histogram_file = not (self.is_hdf5_output)
        if separate_histogram_file:
            fmode = "w"
            hist_fname = path.join(self.output_dir, "histogram_%05d.hdf5" % self.start_index)
        else:
            fmode = "a"
            hist_fname = self.fname
        # Nabu's original NXProcessWriter has to be used here, as histogram is not 3D
        self.histogram_writer = NXProcessWriter(
            hist_fname,
            entry=self._h5_entry,
            filemode=fmode,
            overwrite=True,
        )

    def write_histogram(self, data, config=None, processing_index=1):
        if not (self.histogram):
            return
        self.histogram_writer.write(
            data,
            "histogram",
            processing_index=processing_index,
            config=config,
            is_frames_stack=False,
            direct_access=False,
        )

    def _write_metadata(self):
        self.writer.metadata = self.metadata
        self.writer.save_metadata()
        if self.is_hdf5_output:
            dicttonx(
                self._h5_results_metadata,
                h5file=self.fname,
                h5path=self._h5_entry,
                update_mode="replace",
                mode="a",
            )

    def write_data(self, data):
        self.writer.data = data
        self.writer.save()
        self._write_metadata()



from .. import version
import numpy as np
import h5py
import argparse
import sys
from .cli_configs import CreateDistortionMapHorizontallyMatchedFromPolyConfig
from .utils import parse_params_values
from ..utils import DictToObj
from ..resources.logger import Logger, LoggerOrPrint


def horizontal_match(user_args=None):
    """This application builds  two arrays. Let us call them map_x and map_z. Both are 2D arrays with shape given by (nz, nx).
    These maps are meant to be used to generate a corrected detector image, using them to obtain  the pixel (i,j) of the corrected
    image by interpolating the raw data at position ( map_z(i,j), map_x(i,j) ).

    This map is determined by a user given  polynomial  P(rs) in the radial variable  rs = sqrt( (z-center_z)**2 + (x-center_x)**2 ) / (nx/2)
    where center_z and center_x give the center around which the deformation is centered.

    The perfect  position (zp,xp) , that would be observed on a perfect detector, of a photon observed at pixel (z,x) of the distorted detector is:

       (zp, xp) = (center_z,  center_x) +  P(rs) *  ( z - center_z  , x - center_x )

    The polynomial is given by P(rs) = rs *(1 + c2 * rs**2 + c4 * rs**4)

    The map is rescaled and reshifted so that a perfect match is realised at the borders of a horizontal line passing by the center. This ensures coerence
    with the procedure of pixel size calibration which is performed moving a needle horizontally and reading the motor positions at the extreme positions.

    The maps are written in the target file, creating it as hdf5 file,  in the datasets

        "/coords_source_x"
        "/coords_source_z"

    The URLs of these two maps can be used for the detector correction of type "map_xz"
    in the nabu configuration file as in this example

         [dataset]
         ...
         detector_distortion_correction = map_xz
         detector_distortion_correction_options = map_x="silx:./map_coordinates.h5?path=/coords_source_x" ;  map_z="silx:./map_coordinates.h5?path=/coords_source_z"

    """

    if user_args is None:
        user_args = sys.argv[1:]

    args = DictToObj(
        parse_params_values(
            CreateDistortionMapHorizontallyMatchedFromPolyConfig,
            parser_description=horizontal_match.__doc__,
            program_version="nabu " + version,
            user_args=user_args,
        )
    )
    logger = Logger("horizontal_match", level=args.loglevel, logfile="horizontal_match.log")

    nz, nx = args.nz, args.nx
    center_x, center_z = (args.center_x, args.center_z)

    c4 = args.c4
    c2 = args.c2

    polynomial = np.poly1d([c4, 0, c2, 0, 1, 0.0])
    # change of variable
    cofv = np.poly1d([1.0 / (nx / 2), 0])
    polynomial = nx / 2 * polynomial(cofv)

    left_border = 0 - center_x
    right_border = nx - 1 - center_x

    def get_rescaling_shift(left_border, right_border, polynomial):
        dl = polynomial(left_border)
        dr = polynomial(right_border)

        rescaling = (dr - dl) / (right_border - left_border)
        shift = -left_border * rescaling + dl
        return rescaling, shift

    final_grid_rescaling, final_grid_shift = get_rescaling_shift(left_border, right_border, polynomial)

    coords_z, coords_x = np.indices([nz, nx])

    coords_z = ((coords_z - center_z) * final_grid_rescaling).astype("d")
    coords_x = ((coords_x - center_x) * final_grid_rescaling + final_grid_shift).astype("d")

    distances_goal = np.sqrt(coords_z * coords_z + coords_x * coords_x)

    distances_unknown = distances_goal

    pp_deriv = polynomial.deriv()

    # iteratively finding the positions to interpolated at by newton method
    for i in range(10):
        errors = polynomial(distances_unknown) - distances_goal
        derivative = pp_deriv(distances_unknown)
        distances_unknown = distances_unknown - errors / derivative

    distances_data_sources = distances_unknown

    # avoid 0/0
    distances_data_sources[distances_goal < 1] = 1
    distances_goal[distances_goal < 1] = 1

    coords_source_z = coords_z * distances_data_sources / distances_goal + center_z
    coords_source_x = coords_x * distances_data_sources / distances_goal + center_x

    with h5py.File(args.target_file, "w") as f:
        f["coords_source_x"] = coords_source_x
        f["coords_source_z"] = coords_source_z

    if args.axis_pos is not None:
        coord_axis = args.axis_pos - center_x
        new_pos = (polynomial(coord_axis) - final_grid_shift) / final_grid_rescaling + center_x
        logger.info("New axis position at %e it was previously  %e " % (new_pos, args.axis_pos))

        return new_pos
    else:
        return None



from nabu.pipeline.helical import gridded_accumulator, span_strategy
from nabu.testutils import get_data, __do_long_tests__
import os
import numpy as np
import pytest
from nabu.preproc.ccd import Log, CCDFilter
from nabu.preproc.phase import PaganinPhaseRetrieval
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.pipeline.helical import gridded_accumulator, span_strategy
from nabu.pipeline.helical.weight_balancer import WeightBalancer
from nabu.pipeline.helical.helical_utils import find_mirror_indexes

from nabu.cuda.utils import get_cuda_context, __has_pycuda__, __pycuda_error_msg__, replace_array_memory

if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.pipeline.helical.fbp import BackprojectorHelical as FBPClass


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    # This is a helical dataset derived
    # from "crayon" dataset, using 5 slices and covering 2.5 x 360 angular span
    # in halftomo, with vertical translations.

    #  >>> d=load("small_sparse_helical_dataset.npz")
    #  dd=dict( d.items() )
    #  >>> dd["median_clip_threshold"] = 0.04
    #  >>> savez("retouched_test.npz",**dd)

    helical_dataset = get_data("small_sparse_helical_dataset.npz")

    helical_dataset = dict(list(helical_dataset.items()))
    # the radios, in the dataset file, are stored by swapping angular and x dimension
    # so that the fast running dimension runs over the projections.
    # Due to the sparsity of the dataset, where only an handful of slices
    # has signal, this gives a much better compression even when the axis is translating
    # vertically
    helical_dataset["radios"] = np.array(np.swapaxes(helical_dataset["radios_transposed"], 0, 2))
    del helical_dataset["radios_transposed"]

    # adding members: radios,  dark, flats, z_pix_per_proj, x_pix_per_proj, projection_angles_deg,
    # pixel_size_mm, phase_margin_pix,
    #  weigth_field=weigth_field, double_flat
    dataset_keys = [
        "dark",
        "flats",
        "z_pix_per_proj",
        "x_pix_per_proj",
        "projection_angles_deg",
        "pixel_size_mm",
        "phase_margin_pix",
        "weights_field",
        "double_flat",
        "rotation_axis_position",
        "detector_shape_vh",
        "result_inset",
        "radios",
        # further  parameters, added on top of the test data which was originally made for gridded_accumulator
        "median_clip_threshold",
        "distance_m",
        "energy_kev",
        "delta_beta",
        "pixel_size_m",
        "padding_type",
        "phase_margin_for_pag",
        "rec_reference",
        "ref_tol",
        "ref_start",
        "ref_end",
    ]
    # the test dataset is the original one from the accumulator test
    # plus some patched metadeta information for phase retrieval.
    # The original dataset had phase_margin_pix and had 3 usefule slices.
    # Here, to test phase retrieval, we redefine the phase margin to the maximum
    # that we can do with such a small dataset
    helical_dataset["phase_margin_pix"] = helical_dataset["phase_margin_for_pag"]

    for key in dataset_keys:
        setattr(cls, key, helical_dataset[key])

    cls.padding_type = str(cls.padding_type)
    cls.rotation_axis_position = float(cls.rotation_axis_position)
    cls.rtol_regridded = 1.0e-6

    cls.projection_angles_rad = np.rad2deg(cls.projection_angles_deg)


@pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
@pytest.mark.skipif(not (__has_pycuda__), reason="Needs pycuda for this test")
@pytest.mark.usefixtures("bootstrap")
class TestGriddedAccumulator:
    """
    Test the GriddedAccumulator.
    Rebuilds the sinogram for some selected slices of the crayon dataset
    """

    def test_regridding(self):
        span_info = span_strategy.SpanStrategy(
            z_pix_per_proj=self.z_pix_per_proj,
            x_pix_per_proj=self.x_pix_per_proj,
            detector_shape_vh=self.detector_shape_vh,
            phase_margin_pix=self.phase_margin_pix,
            projection_angles_deg=self.projection_angles_deg,
            require_redundancy=True,
            pixel_size_mm=self.pixel_size_mm,
            logger=None,
        )

        # I would like to reconstruct from feaseable height 15 to feaseable  height 18
        # relatively to the first doable slice in the vertical translation direction
        # I get the heights in the detector frame of the first and of the last

        self.reconstruction_space = gridded_accumulator.get_reconstruction_space(
            span_info=span_info, min_scanwise_z=15, end_scanwise_z=18, phase_margin_pix=self.phase_margin_pix
        )

        chunk_info = span_info.get_chunk_info((self.reconstruction_space.my_z_min, self.reconstruction_space.my_z_end))

        sub_region = (
            self.reconstruction_space.my_z_min - self.phase_margin_pix,
            self.reconstruction_space.my_z_end + self.phase_margin_pix,
        )

        # useful projections
        proj_num_start, proj_num_end = chunk_info.angle_index_span

        # the first of the chunk angular range
        my_first_pnum = proj_num_start

        self.accumulator = gridded_accumulator.GriddedAccumulator(
            gridded_radios=self.reconstruction_space.gridded_radios,
            gridded_weights=self.reconstruction_space.gridded_cumulated_weights,
            diagnostic_radios=self.reconstruction_space.diagnostic_radios,
            diagnostic_weights=self.reconstruction_space.diagnostic_weights,
            diagnostic_angles=self.reconstruction_space.diagnostic_proj_angle,
            dark=self.dark,
            flat_indexes=[0, 7501],
            flats=self.flats,
            weights=self.weights_field,
            double_flat=self.double_flat,
        )

        # splitting in sub ranges of 100 projections
        n_granularity = 100
        pnum_start_list = list(np.arange(proj_num_start, proj_num_end, n_granularity))
        pnum_end_list = pnum_start_list[1:] + [proj_num_end]

        for pnum_start, pnum_end in zip(pnum_start_list, pnum_end_list):
            start_in_chunk = pnum_start - my_first_pnum
            end_in_chunk = pnum_end - my_first_pnum

            self._read_data_and_apply_flats(
                slice(pnum_start, pnum_end), slice(start_in_chunk, end_in_chunk), chunk_info, sub_region, span_info
            )

        res_flatfielded = self.reconstruction_space.gridded_radios / self.reconstruction_space.gridded_cumulated_weights

        # but in real pipeline the radio_shape is obtained from the pipeline get_shape utility method
        self._init_ccd_corrections(res_flatfielded.shape[1:])

        # but in the actual pipeline the argument is not given, and the processed stack is the one internally
        # kept by the pipeline object ( self.gridded_radios in the pipeline )
        self._ccd_corrections(res_flatfielded)

        self._init_phase(res_flatfielded.shape[1:])
        processed_radios = self._retrieve_phase(res_flatfielded)

        self._init_mlog(processed_radios.shape)
        self._take_log(processed_radios)

        top_margin = -self.phase_margin_pix if self.phase_margin_pix else None
        processed_weights = self.reconstruction_space.gridded_cumulated_weights[
            :, self.phase_margin_pix : top_margin, :
        ]

        self._init_weight_balancer()
        self._balance_weights(processed_weights)

        self._init_reconstructor(processed_radios.shape)

        i_slice = 0

        self.d_radios_slim.set(processed_radios[:, i_slice, :])
        self._filter()

        self._apply_weights(i_slice, processed_weights)

        res = self._reconstruct()

        test_slicing = slice(self.ref_start, self.ref_end)
        tested_inset = res[test_slicing, test_slicing]
        assert np.max(np.abs(tested_inset - self.rec_reference)) < self.ref_tol

        # uncomment the four following lines to get the slice image
        # import fabio
        # edf = fabio.edfimage.edfimage()
        # edf.data = res
        # edf.write("reconstructed_slice.edf")

        # put the test here

    def _reconstruct(self):
        axis_corrections = np.zeros_like(self.reconstruction_space.gridded_angles_rad)
        self.reconstruction.set_custom_angles_and_axis_corrections(
            self.reconstruction_space.gridded_angles_rad, axis_corrections
        )

        self.reconstruction.backprojection(self.d_radios_slim, output=self.d_rec_res)

        self.d_rec_res.get(self.rec_res)

        return self.rec_res

    def _apply_weights(self, i_slice, weights):
        """d_radios_slim is on gpu"""
        n_provided_angles = self.d_radios_slim.shape[0]

        for first_angle_index in range(0, n_provided_angles, self.num_weight_radios_per_app):
            end_angle_index = min(n_provided_angles, first_angle_index + self.num_weight_radios_per_app)
            self._d_radios_weights[: end_angle_index - first_angle_index].set(
                weights[first_angle_index:end_angle_index, i_slice]
            )
            self.d_radios_slim[first_angle_index:end_angle_index] *= self._d_radios_weights[
                : end_angle_index - first_angle_index
            ]

    def _filter(self):
        self.mirror_angle_relative_indexes = find_mirror_indexes(self.reconstruction_space.gridded_angles_deg)

        self.reconstruction.sino_filter.filter_sino(
            self.d_radios_slim,
            mirror_indexes=self.mirror_angle_relative_indexes,
            rot_center=self.rotation_axis_position,
            output=self.d_radios_slim,
        )

    def _init_reconstructor(self, processed_radios_shape):
        one_slice_data_shape = processed_radios_shape[:1] + processed_radios_shape[2:]

        self.d_radios_slim = garray.zeros(one_slice_data_shape, np.float32)

        # let's make room for loading chunck of weights without necessarily doubling the memory footprint.
        # The weights will be used to multiplied the d_radios_slim.
        # We will proceed by bunches
        self.num_weight_radios_per_app = 200
        self._d_radios_weights = garray.zeros((self.num_weight_radios_per_app,) + one_slice_data_shape[1:], np.float32)

        pixel_size_cm = self.pixel_size_m * 100

        radio_size_h = processed_radios_shape[-1]

        assert (
            2 * self.rotation_axis_position > radio_size_h
        ), """The code of this test is adapted for HA on the right. This seems to be a case
        of HA on the left because  self.rotation_axis_position={self.rotation_axis_position} and radio_size_h = {radio_size_h}
        """

        rec_dim = int(round(2 * self.rotation_axis_position))

        self.reconstruction = FBPClass(
            one_slice_data_shape,
            angles=np.zeros(processed_radios_shape[0], "f"),
            rot_center=self.rotation_axis_position,
            filter_name=None,
            slice_roi=(0, rec_dim, 0, rec_dim),
            extra_options={
                "scale_factor": 1.0 / pixel_size_cm,
                "axis_correction": np.zeros(processed_radios_shape[0], "f"),
                "padding_mode": "edge",
            },
        )
        self.reconstruction.fbp = self.reconstruction.backproj

        self.d_rec_res = garray.zeros((rec_dim, rec_dim), np.float32)
        self.rec_res = np.zeros((rec_dim, rec_dim), np.float32)

    def _init_weight_balancer(self):
        self.weight_balancer = WeightBalancer(self.rotation_axis_position, self.reconstruction_space.gridded_angles_rad)

    def _balance_weights(self, weights):
        self.weight_balancer.balance_weights(weights)

    def _retrieve_phase(self, radios):
        processed_radios = np.zeros(
            (radios.shape[0],) + (radios.shape[1] - 2 * self.phase_margin_pix,) + (radios.shape[2],), radios.dtype
        )
        for i in range(radios.shape[0]):
            processed_radios[i] = self.phase_retrieval.apply_filter(radios[i])
        return processed_radios

    def _read_data_and_apply_flats(self, sub_total_prange_slice, subchunk_slice, chunk_info, sub_region, span_info):
        my_integer_shifts_v = chunk_info.integer_shift_v[subchunk_slice]

        subr_start_z, subr_end_z = sub_region

        subr_start_z_list = subr_start_z - my_integer_shifts_v
        subr_end_z_list = subr_end_z - my_integer_shifts_v + 1

        dtasrc_start_z = max(0, subr_start_z_list.min())
        dtasrc_end_z = min(span_info.detector_shape_vh[0], subr_end_z_list.max())

        data_raw = self.radios[sub_total_prange_slice, slice(dtasrc_start_z, dtasrc_end_z), :]

        subsampling_file_slice = sub_total_prange_slice

        # my_subsampled_indexes = self.chunk_reader._sorted_files_indices[subsampling_file_slice]
        my_subsampled_indexes = (np.arange(10000))[subsampling_file_slice]

        self.accumulator.extract_preprocess_with_flats(
            subchunk_slice,
            my_subsampled_indexes,
            chunk_info,
            np.array((subr_start_z, subr_end_z), "i"),
            np.array((dtasrc_start_z, dtasrc_end_z), "i"),
            data_raw,
        )

    def _init_ccd_corrections(self, radio_shape):
        # but in real pipeline the radio_shape is obtained from the pipeline get_shape utility method
        self.ccd_correction = CCDFilter(radio_shape, median_clip_thresh=self.median_clip_threshold)

    def _ccd_corrections(self, radios):
        _tmp_radio = np.empty_like(radios[0])
        for i in range(radios.shape[0]):
            self.ccd_correction.median_clip_correction(radios[i], output=_tmp_radio)
            radios[i][:] = _tmp_radio[:]

    def _take_log(self, radios):
        self.mlog.take_logarithm(radios)

    def _init_mlog(self, radios_shape):
        log_shape = radios_shape
        clip_min = 1.0e-6
        clip_max = 1.1
        self.mlog = Log(log_shape, clip_min=clip_min, clip_max=clip_max)

    def _init_phase(self, raw_shape):
        self.phase_retrieval = PaganinPhaseRetrieval(
            raw_shape,
            distance=self.distance_m,
            energy=self.energy_kev,
            delta_beta=self.delta_beta,
            pixel_size=self.pixel_size_m,
            padding=self.padding_type,
            margin=((self.phase_margin_pix,) * 2, (0, 0)),
            use_R2C=True,
            fftw_num_threads=True,  # TODO tune in advanced params of nabu config file
        )
        if self.phase_retrieval.use_fftw:
            self.logger.debug(
                "PaganinPhaseRetrieval using FFTW with %d threads" % self.phase_retrieval.fftw.num_threads
            )



from typing import Union
import numpy as np
import pycuda.gpuarray as garray
from ..preproc.flatfield import FlatFieldArrays
from ..cuda.kernel import CudaKernel
from ..utils import get_cuda_srcfile
from ..io.reader import load_images_from_dataurl_dict


class CudaFlatFieldArrays(FlatFieldArrays):
    def __init__(
        self,
        radios_shape: tuple,
        flats: dict,
        darks: dict,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
        nan_value=1.0,
        radios_srcurrent=None,
        flats_srcurrent=None,
        cuda_options: Union[dict, None] = None,
    ):
        """
        Initialize a flat-field normalization CUDA process.
        Please read the documentation of nabu.preproc.flatfield.FlatField for help
        on the parameters.
        """
        #
        if distortion_correction is not None:
            raise NotImplementedError("Flats distortion correction is not implemented with the Cuda backend")
        #
        super().__init__(
            radios_shape,
            flats,
            darks,
            radios_indices=radios_indices,
            interpolation=interpolation,
            distortion_correction=distortion_correction,
            radios_srcurrent=radios_srcurrent,
            flats_srcurrent=flats_srcurrent,
            nan_value=nan_value,
        )
        self._set_cuda_options(cuda_options)
        self._init_cuda_kernels()
        self._load_flats_and_darks_on_gpu()

    def _set_cuda_options(self, user_cuda_options):
        self.cuda_options = {"device_id": None, "ctx": None, "cleanup_at_exit": None}
        if user_cuda_options is None:
            user_cuda_options = {}
        self.cuda_options.update(user_cuda_options)

    def _init_cuda_kernels(self):
        # TODO
        if self.interpolation != "linear":
            raise ValueError("Interpolation other than linar is not yet implemented in the cuda back-end")
        #
        self._cuda_fname = get_cuda_srcfile("flatfield.cu")
        options = [
            "-DN_FLATS=%d" % self.n_flats,
            "-DN_DARKS=%d" % self.n_darks,
        ]
        if self.nan_value is not None:
            options.append("-DNAN_VALUE=%f" % self.nan_value)
        self.cuda_kernel = CudaKernel(
            "flatfield_normalization", self._cuda_fname, signature="PPPiiiPP", options=options
        )
        self._nx = np.int32(self.shape[1])
        self._ny = np.int32(self.shape[0])

    def _load_flats_and_darks_on_gpu(self):
        # Flats
        self.d_flats = garray.zeros((self.n_flats,) + self.shape, np.float32)
        for i, flat_idx in enumerate(self._sorted_flat_indices):
            self.d_flats[i].set(np.ascontiguousarray(self.flats[flat_idx], dtype=np.float32))
        # Darks
        self.d_darks = garray.zeros((self.n_darks,) + self.shape, np.float32)
        for i, dark_idx in enumerate(self._sorted_dark_indices):
            self.d_darks[i].set(np.ascontiguousarray(self.darks[dark_idx], dtype=np.float32))
        self.d_darks_indices = garray.to_gpu(np.array(self._sorted_dark_indices, dtype=np.int32))
        # Indices
        self.d_flats_indices = garray.to_gpu(self.flats_idx)
        self.d_flats_weights = garray.to_gpu(self.flats_weights)

    def normalize_radios(self, radios):
        """
        Apply a flat-field correction, with the current parameters, to a stack
        of radios.

        Parameters
        -----------
        radios_shape: `pycuda.gpuarray.GPUArray`
            Radios chunk.
        """
        if not (isinstance(radios, garray.GPUArray)):
            raise ValueError("Expected a pycuda.gpuarray (got %s)" % str(type(radios)))
        if radios.dtype != np.float32:
            raise ValueError("radios must be in float32 dtype (got %s)" % str(radios.dtype))
        if radios.shape != self.radios_shape:
            raise ValueError("Expected radios shape = %s but got %s" % (str(self.radios_shape), str(radios.shape)))
        self.cuda_kernel(
            radios,
            self.d_flats,
            self.d_darks,
            self._nx,
            self._ny,
            np.int32(self.n_radios),
            self.d_flats_indices,
            self.d_flats_weights,
        )
        if self.normalize_srcurrent:
            for i in range(self.n_radios):
                radios[i] *= self.srcurrent_ratios[i]
        return radios


CudaFlatField = CudaFlatFieldArrays


class CudaFlatFieldDataUrls(CudaFlatField):
    def __init__(
        self,
        radios_shape: tuple,
        flats: dict,
        darks: dict,
        radios_indices=None,
        interpolation: str = "linear",
        distortion_correction=None,
        nan_value=1.0,
        radios_srcurrent=None,
        flats_srcurrent=None,
        cuda_options: Union[dict, None] = None,
        **chunk_reader_kwargs,
    ):
        flats_arrays_dict = load_images_from_dataurl_dict(flats, **chunk_reader_kwargs)
        darks_arrays_dict = load_images_from_dataurl_dict(darks, **chunk_reader_kwargs)
        super().__init__(
            radios_shape,
            flats_arrays_dict,
            darks_arrays_dict,
            radios_indices=radios_indices,
            interpolation=interpolation,
            distortion_correction=distortion_correction,
            radios_srcurrent=radios_srcurrent,
            flats_srcurrent=flats_srcurrent,
            cuda_options=cuda_options,
        )



import numpy as np
from pycuda import gpuarray as garray
from ..utils import calc_padding_lengths, updiv, get_cuda_srcfile
from ..cuda.processing import CudaProcessing
from ..cuda.kernel import CudaKernel
from ..cuda.padding import CudaPadding
from .phase_cuda import CudaPaganinPhaseRetrieval
from .ctf import CTFPhaseRetrieval


# TODO:
#  - better padding scheme (for now 2*shape)
#  - rework inheritance scheme ? (base class SingleDistancePhaseRetrieval and its cuda counterpart)
class CudaCTFPhaseRetrieval(CTFPhaseRetrieval):

    """
    Cuda back-end of CTFPhaseRetrieval
    """

    def __init__(
        self,
        shape,
        geo_pars,
        delta_beta,
        padded_shape="auto",
        padding_mode="reflect",
        translation_vh=None,
        normalize_by_mean=False,
        lim1=1.0e-5,
        lim2=0.2,
        use_rfft=True,
        fftw_num_threads=None,
        logger=None,
        cuda_options=None,
    ):
        """
        Initialize a CudaCTFPhaseRetrieval.

        Parameters
        ----------
        shape: tuple
            Shape of the images to process
        padding_mode: str
            Padding mode. Default is "reflect".

        Other parameters
        -----------------
        Please refer to CTFPhaseRetrieval documentation.
        """
        if not use_rfft:
            raise ValueError("Only use_rfft=True is supported")
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        super().__init__(
            shape,
            geo_pars,
            delta_beta,
            padded_shape=padded_shape,
            padding_mode=padding_mode,
            translation_vh=translation_vh,
            normalize_by_mean=normalize_by_mean,
            lim1=lim1,
            lim2=lim2,
            logger=logger,
            use_rfft=True,
            fftw_num_threads=False,
        )
        self._init_ctf_filter()
        self._init_cuda_padding()
        self._init_fft()
        self._init_mult_kernel()

    def _init_ctf_filter(self):
        self._mean_scale_factor = self.unreg_filter_denom[0, 0] * np.prod(self.shape_padded)
        self._d_filter_num = garray.to_gpu(self.unreg_filter_denom).astype("f")
        self._d_filter_denom = garray.to_gpu(
            (1.0 / (2 * self.unreg_filter_denom * self.unreg_filter_denom + self.lim)).astype("f")
        )

    def _init_cuda_padding(self):
        pad_width = calc_padding_lengths(self.shape, self.shape_padded)
        # Custom coordinate transform to get directly FFT layout
        R, C = np.indices(self.shape, dtype=np.int32)
        coords_R = np.roll(
            np.pad(R, pad_width, mode=self.padding_mode), (-pad_width[0][0], -pad_width[1][0]), axis=(0, 1)
        )
        coords_C = np.roll(
            np.pad(C, pad_width, mode=self.padding_mode), (-pad_width[0][0], -pad_width[1][0]), axis=(0, 1)
        )
        self.cuda_padding = CudaPadding(
            self.shape,
            (coords_R, coords_C),
            mode=self.padding_mode,
            # propagate cuda options ?
        )

    def _init_fft(self):
        # Import has to be done here, otherwise scikit-cuda creates a cuda/cublas context at import
        from silx.math.fft.cufft import CUFFT

        self.cufft = CUFFT(template=np.zeros(self.shape_padded, dtype="f"))
        self.d_radio_padded = self.cufft.data_in
        self.d_radio_f = self.cufft.data_out

    def _init_mult_kernel(self):
        self.cpxmult_kernel = CudaKernel(
            "CTF_kernel",
            filename=get_cuda_srcfile("ElementOp.cu"),
            signature="PPPfii",
        )
        Nx = np.int32(self.shape_padded[1] // 2 + 1)
        Ny = np.int32(self.shape_padded[0])
        self._cpxmult_kernel_args = [
            self.d_radio_f,
            self._d_filter_num,
            self._d_filter_denom,
            np.float32(self._mean_scale_factor),
            Nx,
            Ny,
        ]
        blk = (32, 32, 1)
        grd = (updiv(Nx, blk[0]), updiv(Ny, blk[1]))
        self._cpxmult_kernel_kwargs = {"grid": grd, "block": blk}

    set_input = CudaPaganinPhaseRetrieval.set_input

    def retrieve_phase(self, image, output=None):
        """
        Perform padding on an image. Please see the documentation of CTFPhaseRetrieval.retrieve_phase().
        """
        self.set_input(image)
        self.cuda_padding.pad(image, output=self.d_radio_padded)
        if self.normalize_by_mean:
            m = garray.sum(self.d_radio_padded).get() / np.prod(self.shape_padded)
            self.d_radio_padded /= m
        self.cufft.fft(self.d_radio_padded, output=self.d_radio_f)
        self.cpxmult_kernel(*self._cpxmult_kernel_args, **self._cpxmult_kernel_kwargs)
        self.cufft.ifft(self.d_radio_f, output=self.d_radio_padded)

        if output is None:
            output = self.cuda_processing.allocate_array("d_output", self.shape)
        output[:, :] = self.d_radio_padded[: self.shape[0], : self.shape[1]]
        return output



from ..cuda.utils import __has_pycuda__
from ..cuda.convolution import Convolution
from ..cuda.processing import CudaProcessing
from .unsharp import UnsharpMask

if __has_pycuda__:
    from pycuda.elementwise import ElementwiseKernel


class CudaUnsharpMask(UnsharpMask):
    def __init__(self, shape, sigma, coeff, mode="reflect", method="gaussian", cuda_options=None):
        """
        Unsharp Mask, cuda backend.
        """
        super().__init__(shape, sigma, coeff, mode=mode, method=method)
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._init_convolution()
        self._init_mad_kernel()
        self.cuda_processing.init_arrays_to_none(["_d_out"])

    def _init_convolution(self):
        self.convolution = Convolution(
            self.shape,
            self._gaussian_kernel,
            mode=self.mode,
            extra_options={  # Use the lowest amount of memory
                "allocate_input_array": False,
                "allocate_output_array": False,
                "allocate_tmp_array": True,
            },
        )

    def _init_mad_kernel(self):
        # garray.GPUArray.mul_add is out of place...
        self.mad_kernel = ElementwiseKernel(
            "float* array, float fac, float* other, float otherfac",
            "array[i] = fac * array[i] + otherfac * other[i]",
            name="mul_add",
        )

    def unsharp(self, image, output=None):
        if output is None:
            output = self.cuda_processing.allocate_array("_d_out", self.shape, "f")
        self.convolution(image, output=output)
        if self.method == "gaussian":
            self.mad_kernel(output, -self.coeff, image, 1.0 + self.coeff)
        elif self.method == "log":
            # output = output * coeff + image   where output was image_blurred
            self.mad_kernel(output, self.coeff, image, 1.0)
        else:  # "imagej":
            # output = (image - coeff*image_blurred)/(1-coeff)  where output was image_blurred
            self.mad_kernel(output, -self.coeff / (1 - self.coeff), image, 1.0 / (1 - self.coeff))
        return output



import numpy as np
from silx.opencl.backprojection import Backprojection
from ..utils import deprecation_warning


# Compatibility layer Nabu/silx
class Backprojector:
    def __init__(
        self,
        sino_shape,
        slice_shape=None,
        angles=None,
        rot_center=None,
        filter_name=None,
        padding_mode=None,
        slice_roi=None,
        scale_factor=None,
        ctx=None,
        devicetype="all",
        platformid=None,
        deviceid=None,
        profile=False,
        extra_options=None,
    ):
        if slice_roi and (
            slice_roi[0] > 0 or slice_roi[2] > 0 or slice_roi[1] < sino_shape[1] or slice_roi[3] < sino_shape[1]
        ):
            raise ValueError("Not implemented yet in the OpenCL back-end")

        self._configure_extra_options(extra_options, padding_mode)
        self._get_scale_factor(scale_factor)

        self.backprojector = Backprojection(
            sino_shape,
            slice_shape=slice_shape,
            axis_position=rot_center,  #
            angles=angles,
            filter_name=filter_name,
            ctx=ctx,
            devicetype=devicetype,
            platformid=platformid,
            deviceid=deviceid,
            profile=profile,
            extra_options=self._silx_fbp_extra_options,
        )

    def _configure_extra_options(self, extra_options, padding_mode):
        self.extra_options = extra_options or {}
        self._silx_fbp_extra_options = {}
        if "padding_mode" in self.extra_options:
            deprecation_warning(
                "Please use 'padding_mode' directly in Backprojector arguments, not in 'extra_options'",
                do_print=True,
                func_name="ocl_fbp_padding_mode",
            )
        if self.extra_options.get("clip_outer_circle", False) or self.extra_options.get("center_slice", False):
            raise NotImplementedError()
        if padding_mode is not None:
            self._silx_fbp_extra_options["padding_mode"] = padding_mode
        self._silx_fbp_extra_options["cutoff"] = self.extra_options.get("fbp_filter_cutoff", 1.0)

    def _get_scale_factor(self, scale_factor):
        if scale_factor is not None:
            deprecation_warning(
                "Please use the 'scale_factor' parameter in extra_options", func_name="ocl_fbp_scale_factor"
            )
        self.scale_factor = scale_factor or self.extra_options.get("scale_factor", None)

    # scale_factor is not implemented in the opencl code
    def _fbp_with_scale_factor(self, sino, output=None):
        return self.backprojector.filtered_backprojection(sino * self.scale_factor, output=output)

    def _fbp(self, sino, output=None):
        return self.backprojector.filtered_backprojection(sino, output=output)

    def filtered_backprojection(self, sino, output=None):
        input_sino = sino
        # TODO scale_factor is not implemented in the silx opencl code
        # This makes a copy of the input array
        if self.scale_factor is not None:
            input_sino = sino * self.scale_factor
        #
        if output is None or isinstance(output, np.ndarray):
            res = self.backprojector.filtered_backprojection(input_sino)
            if output is not None:
                output[:] = res[:]
                return output
            return res
        else:  # assuming pyopencl array
            return self.backprojector.filtered_backprojection(input_sino, output=output)

    fbp = filtered_backprojection

    def backproj(self, *args, **kwargs):
        # TODO scale_factor ?
        return self.backprojector.backprojection(*args, **kwargs)



import h5py
import numpy as np
from scipy.special import erf  # pylint: disable=all
from scipy.special import erf
import sys
import os
from scipy.ndimage import gaussian_filter
from tomoscan.esrf.scan.hdf5scan import ImageKey, HDF5TomoScan
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer
from ..io.reader import load_images_from_dataurl_dict


def main():
    """auxiliary program that can be called to create  default input detector profiles, for nabu helical,
    concerning the weights of the pixels and the "double flat" renormalisation denominator.
    The result is an hdf5 file that can be used as a "processes_file" in the nabu configuration and is used by nabu-helical.
    In particulars cases the user may have fancy masks and correction map and will provide its own processes file,
    and will not need this.

    This code, and in particular the auxiliary function below (that by the way tomwer can use)
    provide a default construction of such maps. The double-flat is set to one and
    the weight is build on the basis of the flat fields   from the dataset with an apodisation on the borders
    which allows to eliminate discontinuities in the contributions from the borders, above and below
    for the z-translations, and on the left or roght border for half-tomo.

    The usage is ::

       nabu-helical-prepare-weights-double nexus_file_name entry_name

    Then the resulting file can be used as processes file in the configuration file of nabu-helical

    """
    if len(sys.argv) not in [3, 4, 5]:
        message = f""" Usage:
            {os.path.basename(sys.argv[0])}   nexus_file_name entry_name [target_file name [transition_width]]
        """
        print(message)
        sys.exit(1)

    file_name = sys.argv[1]
    if len(os.path.dirname(file_name)) == 0:
        # To make sure that other utility routines can succesfully deal with path within the current directory
        file_name = os.path.join(".", file_name)

    entry_name = sys.argv[2]
    process_file_name = "double.h5"
    transition_width = 50.0

    if len(sys.argv) in [4, 5]:
        process_file_name = sys.argv[3]
        if len(sys.argv) in [5]:
            transition_width = float(sys.argv[4])

    dataset_info = HDF5DatasetAnalyzer(file_name, extra_options={"h5_entry": entry_name})
    update_dataset_info_flats_darks(dataset_info, flatfield_mode=1)

    mappe = 0
    my_flats = load_images_from_dataurl_dict(dataset_info.flats)

    for key, flat in my_flats.items():
        mappe += flat
    mappe = mappe / len(list(dataset_info.flats.keys()))

    create_heli_maps(mappe, process_file_name, entry_name, transition_width)


def create_heli_maps(profile, process_file_name, entry_name, transition_width):
    profile = profile / profile.max()
    profile = profile.astype("f")

    profile = gaussian_filter(profile, 10)

    if os.path.exists(process_file_name):
        fd = h5py.File(process_file_name, "r+")
    else:
        fd = h5py.File(process_file_name, "w")

    def f(L, m, w):
        x = np.arange(L)

        d = (x - L + m).astype("f")
        res_r = (1 - erf(d / w)) / 2

        d = (x - m).astype("f")
        res_l = (1 + erf(d / w)) / 2

        return res_r * res_l

    border = f(profile.shape[1], 20, 13.33)

    border_v = f(profile.shape[0], int(round(transition_width / 2)), transition_width / 4)

    path_weights = entry_name + "/weights_field/results/data"
    path_double = entry_name + "/double_flatfield/results/data"

    if path_weights in fd:
        del fd[path_weights]
    if path_double in fd:
        del fd[path_double]

    fd[path_weights] = (profile * border) * border_v[:, None]
    fd[path_double] = np.ones_like(profile)



from os import path
from time import time
from math import ceil
import numpy as np
from silx.io.url import DataUrl
from ...utils import remove_items_from_list
from ...resources.logger import LoggerOrPrint
from ...resources.utils import extract_parameters
from ...io.reader import ChunkReader, HDF5Loader
from ...preproc.ccd import Log, CCDFilter
from ...preproc.flatfield import FlatFieldDataUrls
from ...preproc.distortion import DistortionCorrection
from ...preproc.shift import VerticalShift
from ...preproc.double_flatfield import DoubleFlatField
from ...preproc.phase import PaganinPhaseRetrieval
from ...preproc.ctf import CTFPhaseRetrieval, GeoPars
from ...reconstruction.sinogram import SinoBuilder, SinoNormalization
from ...misc.rotation import Rotation
from ...reconstruction.rings import MunchDeringer
from ...misc.unsharp import UnsharpMask
from ...misc.histogram import PartialHistogram, hist_as_2Darray
from ..utils import use_options, pipeline_step, get_subregion
from ..datadump import DataDumpManager
from ..writer import WriterManager
from ..detector_distortion_provider import DetectorDistortionProvider

# For now we don't have a plain python/numpy backend for reconstruction
try:
    from ...reconstruction.fbp_opencl import Backprojector
except:
    Backprojector = None


class ChunkedPipeline:
    """
    Pipeline for "regular" full-field tomography.
    Data is processed by chunks. A chunk consists in K contiguous lines of all the radios.
    In parallel geometry, a chunk of K radios lines gives K sinograms,
    and equivalently K reconstructed slices.
    """

    backend = "numpy"
    FlatFieldClass = FlatFieldDataUrls
    DoubleFlatFieldClass = DoubleFlatField
    CCDCorrectionClass = CCDFilter
    PaganinPhaseRetrievalClass = PaganinPhaseRetrieval
    CTFPhaseRetrievalClass = CTFPhaseRetrieval
    UnsharpMaskClass = UnsharpMask
    ImageRotationClass = Rotation
    VerticalShiftClass = VerticalShift
    SinoDeringerClass = MunchDeringer
    MLogClass = Log
    SinoBuilderClass = SinoBuilder
    SinoNormalizationClass = SinoNormalization
    FBPClass = Backprojector
    ConebeamClass = None  # unsupported on CPU
    HistogramClass = PartialHistogram

    _default_extra_options = {}

    # These steps are skipped if the reconstruction is done in two stages.
    # The first stage will skip these steps, and the second stage will do these stages after merging sinograms.
    _reconstruction_steps = ["sino_rings_correction", "reconstruction", "save", "histogram"]

    def __init__(
        self, process_config, chunk_shape, margin=None, logger=None, use_grouped_mode=False, extra_options=None
    ):
        """
        Initialize a "Chunked" pipeline.

        Parameters
        ----------
        processing_config: `ProcessConfig`
            Process configuration.
        chunk_shape: tuple
            Shape of the chunk of data to process, in the form (n_angles, n_z, n_x).
            It has to account for possible cropping of the data, eg. [:, start_z:end_z, start_x:end_x]
            where start_xz and/or end_xz can be other than None.
        margin: tuple, optional
            Margin to use, in the form ((up, down), (left, right)).
            It is used for example when performing phase retrieval or a convolution-like operation:
            some extra data is kept to avoid boundaries issues.
            These boundaries are then discarded: the data volume is eventually cropped as
            `data[U:D, L:R]` where `((U, D), (L, R)) = margin`
            If not provided, no margin is applied.
        logger: `nabu.app.logger.Logger`, optional
            Logger class
        extra_options: dict, optional
            Advanced extra options.


        Notes
        ------

        Using `margin` results in a lesser number of reconstructed slices.
        More specifically, if `margin = (V, H)`, then there will be `delta_z - 2*V`
        reconstructed slices (if the sub-region is in the middle of the volume)
        or `delta_z - V` reconstructed slices (if the sub-region is on top or bottom
        of the volume).
        """
        self.logger = LoggerOrPrint(logger)
        self._set_params(process_config, chunk_shape, extra_options, margin, use_grouped_mode)
        self._init_pipeline()

    def _set_params(self, process_config, chunk_shape, extra_options, margin, use_grouped_mode):
        self.process_config = process_config
        self.dataset_info = self.process_config.dataset_info
        self.processing_steps = self.process_config.processing_steps.copy()
        self.processing_options = self.process_config.processing_options
        self._set_chunk_shape(chunk_shape, use_grouped_mode)
        self.set_subregion(None)
        self._set_margin(margin)
        self._set_extra_options(extra_options)
        self._callbacks = {}
        self._steps_name2component = {}
        self._steps_component2name = {}

    def _set_chunk_shape(self, chunk_shape, use_grouped_mode):
        if len(chunk_shape) != 3:
            raise ValueError("Expected chunk_shape to be a tuple of length 3 in the form (n_z, n_y, n_x)")
        self.chunk_shape = tuple(int(c) for c in chunk_shape)  # cast to int, as numpy.int64 can make pycuda crash
        # TODO: sanity check (eg. compare to size of radios in dataset_info) ?
        # (n_a, n_z, n_x)
        self.radios_shape = (
            ceil(self.chunk_shape[0] / self.process_config.subsampling_factor),
            self.chunk_shape[1] // self.process_config.binning[1],
            self.chunk_shape[2] // self.process_config.binning[0],
        )
        self.n_angles = self.radios_shape[0]
        self.n_slices = self.radios_shape[1]
        self._grouped_processing = False
        if use_grouped_mode or self.chunk_shape[0] < len(self.process_config.rotation_angles(subsampling=False)):
            # TODO allow a certain tolerance in this case ?
            # Reconstruction is still possible (albeit less accurate) if delta is small
            self._grouped_processing = True
            self.logger.debug("Only a subset of angles is processed - Reconstruction will be skipped")
            self.processing_steps, _ = remove_items_from_list(self.processing_steps, self._reconstruction_steps)

    def _set_margin(self, margin):
        if margin is None:
            U, D, L, R = None, None, None, None
        else:
            ((U, D), (L, R)) = get_subregion(margin, ndim=2)

        # Replace "None" with zeros
        U, D, L, R = U or 0, D or 0, L or 0, R or 0

        self.margin = ((U, D), (L, R))
        self._margin_up = U
        self._margin_down = D
        self._margin_left = L
        self._margin_right = R
        self.use_margin = (U + D + L + R) > 0
        self.n_recs = self.chunk_shape[1] - sum(self.margin[0])
        self.radios_cropped_shape = (self.radios_shape[0], self.radios_shape[1] - U - D, self.radios_shape[2] - L - R)
        if self.use_margin:
            self.n_slices -= sum(self.margin[0])

    def set_subregion(self, sub_region):
        """
        Set the data volume sub-region to process.
        Note that processing margin, if any, is contained within the sub-region.

        Parameters
        -----------
        sub_region: tuple
            Data volume sub-region, in the form ((start_a, end_a), (start_z, end_z), (start_x, end_x))
            where the data volume has a layout (angles, Z, X)
        """
        n_angles = self.dataset_info.n_angles
        n_x, n_z = self.dataset_info.radio_dims
        c_a, c_z, c_x = self.chunk_shape
        if sub_region is None:
            # By default, take the sub-region around central slice
            sub_region = (
                (0, c_a),
                (n_z // 2 - c_z // 2, n_z // 2 - c_z // 2 + c_z),
                (n_x // 2 - c_x // 2, n_x // 2 - c_x // 2 + c_x),
            )
        else:
            sub_region = get_subregion(sub_region, ndim=3)
            # check sub-region
            for i, start_end in enumerate(sub_region):
                start, end = start_end
                if start is not None and end is not None:
                    if end - start != self.chunk_shape[i]:
                        raise ValueError(
                            "Invalid (start, end)=(%d, %d) for sub-region (dimension %d): chunk shape is %s, but %d-%d=%d != %d"
                            % (start, end, i, str(self.chunk_shape), end, start, end - start, self.chunk_shape[i])
                        )
            #
        self.logger.debug("Set sub-region to %s" % (str(sub_region)))
        self.sub_region = sub_region
        self._sub_region_xz = sub_region[2] + sub_region[1]

    def _set_extra_options(self, extra_options):
        self.extra_options = self._default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    #
    # Callbacks
    #

    def register_callback(self, step_name, callback):
        """
        Register a callback for a pipeline processing step.

        Parameters
        ----------
        step_name: str
            processing step name
        callback: callable
            A function. It will be executed once the processing step `step_name`
            is finished. The function takes only one argument: the class instance.
        """
        if step_name not in self.processing_steps:
            raise ValueError("'%s' is not in processing steps %s" % (step_name, self.processing_steps))
        if step_name in self._callbacks:
            self._callbacks[step_name].append(callback)
        else:
            self._callbacks[step_name] = [callback]

    #
    # Memory management
    #

    def _allocate_array(self, shape, dtype, name=None):
        return np.zeros(shape, dtype=dtype)

    def _allocate_recs(self, ny, nx, n_slices=None):
        n_slices = n_slices or self.n_slices
        self.recs = self._allocate_array((n_slices, ny, nx), "f", name="recs")

    #
    # Runtime attributes
    #

    @property
    def sub_region_xz(self):
        """
        Return the currently processed sub-region in the form
        (start_x, end_x, start_z, end_z)
        """
        return self._sub_region_xz

    @property
    def z_min(self):
        return self._sub_region_xz[2]

    @property
    def sino_shape(self):
        return self.process_config.sino_shape(binning=True, subsampling=True, after_ha=True)

    @property
    def sinos_shape(self):
        return (self.n_slices,) + self.sino_shape

    def get_slice_start_index(self):
        return self.z_min + self._margin_up

    #
    # Pipeline initialization
    #

    def _init_pipeline(self):
        self._allocate_radios()
        self._init_data_dump()
        self._init_reader()
        self._init_flatfield()
        self._init_double_flatfield()
        self._init_ccd_corrections()
        self._init_radios_rotation()
        self._init_phase()
        self._init_unsharp()
        self._init_radios_movements()
        self._init_mlog()
        self._init_sino_normalization()
        self._init_sino_rings_correction()
        self._init_reconstruction()
        self._init_histogram()
        self._init_writer()

    def _allocate_radios(self):
        self.radios = np.zeros(self.radios_shape, dtype=np.float32)
        self.data = self.radios  # alias

    def _init_data_dump(self):
        self._resume_from_step = self.processing_options["read_chunk"].get("step_name", None)
        self.datadump_manager = DataDumpManager(
            self.process_config, self.sub_region, margin=self.margin, logger=self.logger
        )
        # When using "grouped processing", sinogram has to be dumped.
        # If it was not specified by user, force sinogram dump
        # Perhaps these lines should be moved directly to DataDumpManager.
        if self._grouped_processing and not self.process_config.dump_sinogram:
            sino_dump_fname = self.process_config.get_save_steps_file("sinogram")
            self.datadump_manager._configure_dump("sinogram", force_dump_to_fname=sino_dump_fname)
            self.logger.debug("Will dump sinogram to %s" % self.datadump_manager.data_dump["sinogram"].fname)

    @use_options("read_chunk", "chunk_reader")
    def _init_reader(self):
        options = self.processing_options["read_chunk"]
        self._update_reader_configuration()

        process_file = options.get("process_file", None)
        if process_file is None:
            # Standard case - start pipeline from raw data
            if self.process_config.nabu_config["preproc"]["detector_distortion_correction"] is None:
                self.detector_corrector = None
            else:
                self.detector_corrector = DetectorDistortionProvider(
                    detector_full_shape_vh=self.process_config.dataset_info.radio_dims[::-1],
                    correction_type=self.process_config.nabu_config["preproc"]["detector_distortion_correction"],
                    options=self.process_config.nabu_config["preproc"]["detector_distortion_correction_options"],
                )
            # ChunkReader always take a non-subsampled dictionary "files".
            self.chunk_reader = ChunkReader(
                self._read_options["files"],
                sub_region=self.sub_region_xz,
                data_buffer=self.radios,
                pre_allocate=False,
                detector_corrector=self.detector_corrector,
                convert_float=True,
                binning=options["binning"],
                dataset_subsampling=options["dataset_subsampling"],
            )
        else:
            # Resume pipeline from dumped intermediate step
            self.chunk_reader = HDF5Loader(
                process_file,
                options["process_h5_path"],
                sub_region=self.datadump_manager.get_read_dump_subregion(),
                data_buffer=self.radios,
                pre_allocate=False,
            )
            self._resume_from_step = options["step_name"]
            self.logger.debug(
                "Load subregion %s from file %s" % (str(self.chunk_reader.sub_region), self.chunk_reader.fname)
            )

    def _update_reader_configuration(self):
        """
        Modify self.processing_options["read_chunk"] to select a subset of the files, if needed
        (i.e when processing only a subset of the images stack)
        """
        self._read_options = self.processing_options["read_chunk"].copy()
        if self.n_angles == self.process_config.n_angles(subsampling=True):
            # Nothing to do if the full angular range is processed in one shot
            return
        if self._resume_from_step is not None:
            if self._resume_from_step == "sinogram":
                msg = "It makes no sense to use 'grouped processing' when resuming from sinogram"
                self.logger.fatal(msg)
                raise ValueError(msg)
            # Nothing to do if we resume the processing from a given step
            return
        input_data_files = {}
        files_indices = sorted(self._read_options["files"].keys())
        angle_idx_start, angle_idx_end = self.sub_region[0]
        for i in range(angle_idx_start, angle_idx_end):
            idx = files_indices[i]
            input_data_files[idx] = self._read_options["files"][idx]
        self._read_options["files"] = input_data_files

    @use_options("flatfield", "flatfield")
    def _init_flatfield(self):
        self._ff_options = self.processing_options["flatfield"].copy()
        # Use chunk_reader.files instead of process_config.projs_indices(subsampling=True), because
        # chunk_reader might read only a subset of the files (in "grouped mode")
        self._ff_options["projs_indices"] = list(self.chunk_reader.files_subsampled.keys())
        if self._ff_options.get("normalize_srcurrent", False):
            a_start_idx, a_end_idx = self.sub_region[0]
            subs = self.process_config.subsampling_factor
            self._ff_options["radios_srcurrent"] = self._ff_options["radios_srcurrent"][a_start_idx:a_end_idx:subs]

        distortion_correction = None
        if self._ff_options["do_flat_distortion"]:
            self.logger.info("Flats distortion correction will be applied")
            self.FlatFieldClass = FlatFieldDataUrls  # no GPU implementation available, force this backend
            estimation_kwargs = {}
            estimation_kwargs.update(self._ff_options["flat_distortion_params"])
            estimation_kwargs["logger"] = self.logger
            distortion_correction = DistortionCorrection(
                estimation_method="fft-correlation", estimation_kwargs=estimation_kwargs, correction_method="interpn"
            )

        # FlatField parameter "radios_indices" must account for subsampling
        self.flatfield = self.FlatFieldClass(
            self.radios_shape,
            flats=self.dataset_info.flats,
            darks=self.dataset_info.darks,
            radios_indices=self._ff_options["projs_indices"],
            interpolation="linear",
            distortion_correction=distortion_correction,
            sub_region=self.sub_region_xz,
            detector_corrector=self.detector_corrector,
            binning=self._ff_options["binning"],
            radios_srcurrent=self._ff_options["radios_srcurrent"],
            flats_srcurrent=self._ff_options["flats_srcurrent"],
            convert_float=True,
        )

    @use_options("double_flatfield", "double_flatfield")
    def _init_double_flatfield(self):
        options = self.processing_options["double_flatfield"]
        avg_is_on_log = options["sigma"] is not None
        result_url = None
        if options["processes_file"] not in (None, ""):
            result_url = DataUrl(
                file_path=options["processes_file"],
                data_path=(self.dataset_info.hdf5_entry or "entry") + "/double_flatfield/results/data",
            )
            self.logger.info("Loading double flatfield from %s" % result_url.file_path())
        if (self.n_angles < self.process_config.n_angles(subsampling=True)) and result_url is None:
            raise ValueError(
                "Cannot use double-flatfield when processing subset of radios. Please use the 'nabu-double-flatfield' command"
            )
        self.double_flatfield = self.DoubleFlatFieldClass(
            self.radios_shape,
            result_url=result_url,
            sub_region=self.sub_region_xz,
            detector_corrector=self.detector_corrector,
            input_is_mlog=False,
            output_is_mlog=False,
            average_is_on_log=avg_is_on_log,
            sigma_filter=options["sigma"],
        )

    @use_options("ccd_correction", "ccd_correction")
    def _init_ccd_corrections(self):
        options = self.processing_options["ccd_correction"]
        self.ccd_correction = self.CCDCorrectionClass(
            self.radios_shape[1:], median_clip_thresh=options["median_clip_thresh"]
        )

    @use_options("rotate_projections", "projs_rot")
    def _init_radios_rotation(self):
        options = self.processing_options["rotate_projections"]
        center = options["center"]
        if center is None:
            nx, ny = self.radios_shape[1:][::-1]  # after binning
            center = (nx / 2 - 0.5, ny / 2 - 0.5)
        center = (center[0], center[1] - self.z_min)
        self.projs_rot = self.ImageRotationClass(
            self.radios_shape[1:], options["angle"], center=center, mode="edge", reshape=False
        )
        self._tmp_rotated_radio = self._allocate_array(self.radios_shape[1:], "f", name="tmp_rotated_radio")

    @use_options("radios_movements", "radios_movements")
    def _init_radios_movements(self):
        options = self.processing_options["radios_movements"]
        self._vertical_shifts = options["translation_movements"][:, 1]
        self.radios_movements = self.VerticalShiftClass(self.radios.shape, self._vertical_shifts)

    @use_options("phase", "phase_retrieval")
    def _init_phase(self):
        options = self.processing_options["phase"]
        if options["method"] == "CTF":
            translations_vh = getattr(self.dataset_info, "ctf_translations", None)
            geo_pars_params = options["ctf_geo_pars"].copy()
            geo_pars_params["logger"] = self.logger
            geo_pars = GeoPars(**geo_pars_params)
            self.phase_retrieval = self.CTFPhaseRetrievalClass(
                self.radios_shape[1:],
                geo_pars,
                options["delta_beta"],
                lim1=options["ctf_lim1"],
                lim2=options["ctf_lim2"],
                logger=self.logger,
                fftw_num_threads=None,  # TODO tune in advanced params of nabu config file
                use_rfft=True,
                normalize_by_mean=options["ctf_normalize_by_mean"],
                translation_vh=translations_vh,
            )
        else:
            self.phase_retrieval = self.PaganinPhaseRetrievalClass(
                self.radios_shape[1:],
                distance=options["distance_m"],
                energy=options["energy_kev"],
                delta_beta=options["delta_beta"],
                pixel_size=options["pixel_size_m"],
                padding=options["padding_type"],
                # TODO tune in advanced params of nabu config file
                fftw_num_threads=None,
            )
        if self.phase_retrieval.use_fftw:
            self.logger.debug(
                "%s using FFTW with %d threads"
                % (self.phase_retrieval.__class__.__name__, self.phase_retrieval.fftw.num_threads)
            )

    @use_options("unsharp_mask", "unsharp_mask")
    def _init_unsharp(self):
        options = self.processing_options["unsharp_mask"]
        self.unsharp_mask = self.UnsharpMaskClass(
            self.radios_shape[1:],
            options["unsharp_sigma"],
            options["unsharp_coeff"],
            mode="reflect",
            method=options["unsharp_method"],
        )

    @use_options("take_log", "mlog")
    def _init_mlog(self):
        options = self.processing_options["take_log"]
        self.mlog = self.MLogClass(
            self.radios_shape, clip_min=options["log_min_clip"], clip_max=options["log_max_clip"]
        )

    @use_options("sino_normalization", "sino_normalization")
    def _init_sino_normalization(self):
        options = self.processing_options["sino_normalization"]
        self.sino_normalization = self.SinoNormalizationClass(
            kind=options["method"],
            radios_shape=self.radios_cropped_shape,
            normalization_array=options["normalization_array"],
        )

    @use_options("sino_rings_correction", "sino_deringer")
    def _init_sino_rings_correction(self):
        options = self.processing_options["sino_rings_correction"]
        fw_params = extract_parameters(options["user_options"])
        fw_sigma = fw_params.pop("sigma", 1.0)
        self.sino_deringer = self.SinoDeringerClass(fw_sigma, sinos_shape=self.sinos_shape, **fw_params)

    @use_options("reconstruction", "reconstruction")
    def _init_reconstruction(self):
        options = self.processing_options["reconstruction"]
        if options["method"] == "FBP" and self.FBPClass is None:
            raise ValueError("No usable FBP module was found")
        if options["method"] == "cone" and self.ConebeamClass is None:
            raise ValueError("No usable cone-beam module was found")

        if options["method"] == "FBP":
            n_slices = self.n_slices
            radios_shape_for_sino_builder = self.radios_cropped_shape
            self.reconstruction = self.FBPClass(
                self.sinos_shape[1:],
                angles=options["angles"],
                rot_center=options["fbp_rotation_axis_position"],
                filter_name=options["fbp_filter_type"],
                slice_roi=self.process_config.rec_roi,
                padding_mode=options["padding_type"],
                extra_options={
                    "scale_factor": 1.0 / options["pixel_size_cm"],
                    "axis_correction": options["axis_correction"],
                    "centered_axis": options["centered_axis"],
                    "clip_outer_circle": options["clip_outer_circle"],
                    "filter_cutoff": options["fbp_filter_cutoff"],
                },
            )
            if options["fbp_filter_type"] is None:
                self.reconstruction.fbp = self.reconstruction.backproj
        if options["method"] == "cone":
            radios_shape_for_sino_builder = self.radios_shape
            n_slices = self.n_slices + sum(self.margin[0])
            # For numerical stability, normalize all lengths with respect to detector pixel size
            pixel_size_m = self.dataset_info.pixel_size * 1e-6
            source_sample_dist = options["source_sample_dist"] / pixel_size_m
            sample_detector_dist = options["sample_detector_dist"] / pixel_size_m
            self.reconstruction = self.ConebeamClass(  # pylint: disable=E1102
                (self.radios_shape[1],) + self.sino_shape,
                source_sample_dist,
                sample_detector_dist,
                angles=options["angles"],
                # TODO one center for each angle to handle "x translations"
                rot_center=options["rotation_axis_position"],
                pixel_size=1,
            )
        self.sino_builder = self.SinoBuilderClass(
            radios_shape=radios_shape_for_sino_builder,
            rot_center=options["rotation_axis_position"],
            halftomo=options["enable_halftomo"],
            angles=options["angles"],
        )
        self._allocate_recs(*self.process_config.rec_shape, n_slices=n_slices)
        if options["method"] == "cone":
            self.sinos = self._allocate_array(self.sino_builder.output_shape, "f", name="sinos")

    @use_options("histogram", "histogram")
    def _init_histogram(self):
        options = self.processing_options["histogram"]
        self.histogram = self.HistogramClass(method="fixed_bins_number", num_bins=options["histogram_bins"])

    @use_options("save", "writer")
    def _init_writer(self):
        options = self.processing_options["save"]
        metadata = {
            "process_name": "reconstruction",
            "processing_index": 0,
            # TODO this one takes too much time to write, not useful for partial files
            # "processing_options": self.processing_options,
            #
            "nabu_config": self.process_config.nabu_config,
            "entry": getattr(self.dataset_info.dataset_scanner, "entry", "entry"),
        }
        writer_extra_options = {
            "jpeg2000_compression_ratio": options["jpeg2000_compression_ratio"],
            "float_clip_values": options["float_clip_values"],
            "tiff_single_file": options.get("tiff_single_file", False),
            "single_output_file_initialized": getattr(self.process_config, "single_output_file_initialized", False),
        }
        self.writer = WriterManager(
            options["location"],
            options["file_prefix"],
            file_format=options["file_format"],
            overwrite=options["overwrite"],
            start_index=self.get_slice_start_index(),
            logger=self.logger,
            metadata=metadata,
            histogram=("histogram" in self.processing_steps),
            extra_options=writer_extra_options,
        )

    #
    # Pipeline execution
    #

    @pipeline_step("chunk_reader", "Reading data")
    def _read_data(self):
        self.logger.debug("Region = %s" % str(self.sub_region))
        t0 = time()
        self.chunk_reader.load_data()
        el = time() - t0
        shp = self.chunk_reader.data.shape
        self.logger.info("Read subvolume %s in %.2f s" % (str(shp), el))

    @pipeline_step("flatfield", "Applying flat-field")
    def _flatfield(self):
        self.flatfield.normalize_radios(self.radios)

    @pipeline_step("double_flatfield", "Applying double flat-field")
    def _double_flatfield(self, radios=None):
        if radios is None:
            radios = self.radios
        self.double_flatfield.apply_double_flatfield(radios)

    @pipeline_step("ccd_correction", "Applying CCD corrections")
    def _ccd_corrections(self, radios=None):
        if radios is None:
            radios = self.radios
        _tmp_radio = self._allocate_array(radios.shape[1:], "f", name="tmp_ccdcorr_radio")
        for i in range(radios.shape[0]):
            self.ccd_correction.median_clip_correction(radios[i], output=_tmp_radio)
            radios[i][:] = _tmp_radio[:]

    @pipeline_step("projs_rot", "Rotating projections")
    def _rotate_projections(self, radios=None):
        if radios is None:
            radios = self.radios
        tmp_radio = self._tmp_rotated_radio
        for i in range(radios.shape[0]):
            self.projs_rot.rotate(radios[i], output=tmp_radio)
            radios[i][:] = tmp_radio[:]

    @pipeline_step("phase_retrieval", "Performing phase retrieval")
    def _retrieve_phase(self):
        for i in range(self.radios.shape[0]):
            self.phase_retrieval.retrieve_phase(self.radios[i], output=self.radios[i])

    @pipeline_step("unsharp_mask", "Performing unsharp mask")
    def _apply_unsharp(self):
        for i in range(self.radios.shape[0]):
            self.radios[i] = self.unsharp_mask.unsharp(self.radios[i])

    @pipeline_step("mlog", "Taking logarithm")
    def _take_log(self):
        self.mlog.take_logarithm(self.radios)

    @pipeline_step("radios_movements", "Applying radios movements")
    def _radios_movements(self, radios=None):
        if radios is None:
            radios = self.radios
        self.radios_movements.apply_vertical_shifts(radios, list(range(radios.shape[0])))

    def _crop_radios(self):
        if self.use_margin:
            self._orig_radios = self.radios
            if self.processing_options.get("reconstruction", {}).get("method", None) == "cone":
                return
            ((U, D), (L, R)) = self.margin
            self.logger.debug(
                "Cropping radios from %s to %s" % (str(self.radios_shape), str(self.radios_cropped_shape))
            )
            U, D, L, R = U or None, -D or None, L or None, -R or None
            self.radios = self.radios[:, U:D, L:R]  # view

    @pipeline_step("sino_normalization", "Normalizing sinograms")
    def _normalize_sinos(self, radios=None):
        if radios is None:
            radios = self.radios
        sinos = radios.transpose((1, 0, 2))
        self.sino_normalization.normalize(sinos)

    def _dump_sinogram(self):
        if self.datadump_manager is not None:
            self.datadump_manager.dump_data_to_file("sinogram", self.radios)

    @pipeline_step("sino_deringer", "Removing rings on sinograms")
    def _destripe_sinos(self):
        sinos = np.rollaxis(self.radios, 1, 0)  # view
        self.sino_deringer.remove_rings(sinos)  # TODO check it works with non-contiguous view

    @pipeline_step("reconstruction", "Reconstruction")
    def _reconstruct(self):
        """
        Reconstruction for parallel geometry.
        For each target slice: get the corresponding sinogram, apply some processing, then reconstruct
        """
        if self.processing_options["reconstruction"]["method"] == "cone":
            self._reconstruct_cone()
            return
        for i in range(self.n_slices):
            sino = self.sino_builder.get_sino(self.radios, i)
            if self.sino_deringer is not None:
                self.sino_deringer.remove_rings(sino)
            self.reconstruction.fbp(sino, output=self.recs[i])

    def _reconstruct_cone(self):
        """
        This reconstructs the entire sinograms stack at once
        """
        self.sino_builder.get_sinos(self.radios, output=self.sinos)
        z_min, z_max = self.sub_region_xz[2:]
        n_z = self.process_config.radio_shape(binning=True)[0]

        self.reconstruction.reconstruct(  # pylint: disable=E1101
            self.sinos,
            output=self.recs,
            relative_z_position=((z_min + z_max) / self.process_config.binning_z / 2) - n_z / 2,
        )

    @pipeline_step("histogram", "Computing histogram")
    def _compute_histogram(self, data=None):
        if data is None:
            data = self.recs
        self.recs_histogram = self.histogram.compute_histogram(data)

    @pipeline_step("writer", "Saving data")
    def _write_data(self, data=None):
        if data is None:
            data = self.recs
        self.writer.write_data(data)
        self.logger.info("Wrote %s" % self.writer.fname)
        self._write_histogram()
        self.process_config.single_output_file_initialized = True

    def _write_histogram(self):
        if "histogram" not in self.processing_steps:
            return
        self.logger.info("Saving histogram")
        self.writer.write_histogram(
            hist_as_2Darray(self.recs_histogram),
            processing_index=1,
            config={
                "file": path.basename(self.writer.fname),
                "bins": self.processing_options["histogram"]["histogram_bins"],
            },
        )

    def _process_finalize(self):
        if self.use_margin:
            self.radios = self._orig_radios

    def __repr__(self):
        res = "%s(%s, margin=%s)" % (self.__class__.__name__, str(self.chunk_shape), str(self.margin))
        binning = self.process_config.binning
        subsampling = self.process_config.subsampling_factor

        if binning != (1, 1) or subsampling > 1:
            if binning != (1, 1):
                res += "\nImages binning: %s" % (str(binning))
            if subsampling:
                res += "\nAngles subsampling: %d" % subsampling
            res += "\nRadios chunk: %s ---> %s" % (self.chunk_shape, self.radios_shape)

        if self.use_margin:
            res += "\nMargin: %s" % (str(self.margin))
            res += "\nRadios chunk: %s ---> %s" % (str(self.radios_shape), str(self.radios_cropped_shape))

        res += "\nCurrent subregion: %s" % (str(self.sub_region))
        for step_name in self.processing_steps:
            res += "\n- %s" % (step_name)
        return res

    def _process_chunk(self):
        self._flatfield()
        self._double_flatfield()
        self._ccd_corrections()
        self._rotate_projections()
        self._retrieve_phase()
        self._apply_unsharp()
        self._take_log()
        self._radios_movements()
        self._crop_radios()
        self._normalize_sinos()
        self._dump_sinogram()
        # self._destripe_sinos()
        self._reconstruct()
        self._compute_histogram()
        self._write_data()
        self._process_finalize()

    def _reset_reader_subregion(self):
        if self._resume_from_step is None:
            # Normal mode - read data from raw radios
            self.chunk_reader._set_subregion(self.sub_region_xz)
            self.chunk_reader._init_reader()
            self.chunk_reader._loaded = False
        else:
            # Resume from a checkpoint. In this case, we have to re-initialize "datadump manager"
            # sooner to configure start_xyz, end_xyz
            self._init_data_dump()
            self.chunk_reader._set_subregion(self.datadump_manager.get_read_dump_subregion())
            self.chunk_reader._loaded = False
        if self._grouped_processing:
            self._update_reader_configuration()
            self.chunk_reader._set_files(self._read_options["files"])

    def _reset_sub_region(self, sub_region):
        self.set_subregion(sub_region)
        # When sub_region is changed, all components involving files reading have to be updated
        self._reset_reader_subregion()
        self._init_flatfield()  # reset flatfield
        self._init_writer()
        self._init_double_flatfield()
        self._init_data_dump()

    def process_chunk(self, sub_region):
        self._reset_sub_region(sub_region)
        self._read_data()
        self._process_chunk()



from ..config_validators import *

nabu_config = {
    "dataset": {
        "location": {
            "default": "",
            "help": "Dataset location, either a directory or a HDF5-Nexus file.",
            "validator": dataset_location_validator,
            "type": "required",
        },
        "hdf5_entry": {
            "default": "",
            "help": "Entry in the HDF5 file, if applicable. Default is the first available entry.",
            "validator": optional_string_validator,
            "type": "advanced",
        },
        "nexus_version": {
            "default": "1.0",
            "help": "Nexus version to use when browsing the HDF5 dataset. Default is 1.0.",
            "validator": float_validator,
            "type": "advanced",
        },
        "darks_flats_dir": {
            "default": "",
            "help": "Path to a directory where XXX_flats.h5 and XXX_darks.h5 are to be found, where 'XXX' denotes the dataset basename. If these files are found, then reduced flats/darks will be loaded from them. Otherwise, reduced flats/darks will be saved to there once computed, either in the .nx directory, or in the output directory. Mind that the HDF5 entry corresponds to the one of the dataset.",
            "validator": optional_directory_location_validator,
            "type": "optional",
        },
        "binning": {
            "default": "1",
            "help": "Binning factor in the horizontal dimension when reading the data.\nThe final slices dimensions will be divided by this factor.",
            "validator": binning_validator,
            "type": "advanced",
        },
        "binning_z": {
            "default": "1",
            "help": "Binning factor in the vertical dimension when reading the data.\nThis results in a lesser number of reconstructed slices.",
            "validator": binning_validator,
            "type": "advanced",
        },
        "projections_subsampling": {
            "default": "1",
            "help": "Projections subsampling factor: take one projection out of 'projection_subsampling'",
            "validator": binning_validator,
            "type": "advanced",
        },
        "exclude_projections": {
            "default": "",
            "help": "Path to a file name containing projections to exclude (projection indices).",
            "validator": optional_file_location_validator,
            "type": "advanced",
        },
        "overwrite_metadata": {
            "default": "",
            "help": "Which metadata to overwrite, separated by a semicolon, and with units. Example: 'energy = 19 kev; pixel_size = 1.6 um'",
            "validator": no_validator,
            "type": "advanced",
        },
    },
    "preproc": {
        "flatfield": {
            "default": "1",
            "help": "How to perform flat-field normalization. The parameter value can be:\n - 1 or True: enabled.\n - 0 or False: disabled\n - forced or force-load: perform flatfield regardless of the dataset by attempting to load darks/flats\n - force-compute: perform flatfield, ignore all .h5 files containing already computed darks/flats.",
            "validator": flatfield_enabled_validator,
            "type": "required",
        },
        "flat_distortion_correction_enabled": {
            "default": "0",
            "help": "Whether to correct for flat distortion. If activated, each radio is correlated with its corresponding flat, in order to determine and correct the flat distortion.",
            "validator": boolean_validator,
            "type": "advanced",
        },
        "flat_distortion_params": {
            "default": "tile_size=100; interpolation_kind='linear'; padding_mode='edge'; correction_spike_threshold=None",
            "help": "Advanced parameters for flat distortion correction",
            "validator": optional_string_validator,
            "type": "advanced",
        },
        "normalize_srcurrent": {
            "default": "0",
            "help": "Whether to normalize frames with Synchrotron Current. This can correct the effect of a beam refill not taken into account by flats.",
            "validator": boolean_validator,
            "type": "advanced",
        },
        "ccd_filter_enabled": {
            "default": "0",
            "help": "Whether to enable the CCD hotspots correction.",
            "validator": boolean_validator,
            "type": "optional",
        },
        "ccd_filter_threshold": {
            "default": "0.04",
            "help": "If ccd_filter_enabled = 1, a median filter is applied on the 3X3 neighborhood\nof every pixel. If a pixel value exceeds the median value more than this parameter,\nthen the pixel value is replaced with the median value.",
            "validator": float_validator,
            "type": "optional",
        },
        "detector_distortion_correction": {
            "default": "",
            "help": "Apply coordinate transformation on the raw data, at the reading stage. Default (empty) is None. Available are: None, identity(for testing the pipeline), map_xz. This latter method requires two URLs being passet by detector_distortion_correction_options: map_x and map_z pointing to two 2D arrays containing the position where each pixel can be interpolated at in the raw data",
            "validator": detector_distortion_correction_validator,
            "type": "advanced",
        },
        "detector_distortion_correction_options": {
            "default": "",
            "help": "Options for sinogram rings correction methods. The parameters are separated by commas and passed as 'name=value', for example: center_xz=(1000,100); angle_deg=5. Mind the semicolon separator (;).",
            "validator": generic_options_validator,
            "type": "advanced",
        },
        "double_flatfield_enabled": {
            "default": "0",
            "help": "Whether to enable the 'double flat-field' filetering for correcting rings artefacts.",
            "validator": boolean_validator,
            "type": "optional",
        },
        "dff_sigma": {
            "default": "",
            "help": "Enable high-pass filtering on double flatfield with this value of 'sigma'",
            "validator": optional_float_validator,
            "type": "advanced",
        },
        "take_logarithm": {
            "default": "1",
            "help": "Whether to take logarithm after flat-field and phase retrieval.",
            "validator": boolean_validator,
            "type": "required",
        },
        "log_min_clip": {
            "default": "1e-6",
            "help": "After division by the FF, and before the logarithm, the is clipped to this minimum. Enabled only if take_logarithm=1",
            "validator": float_validator,
            "type": "advanced",
        },
        "log_max_clip": {
            "default": "10.0",
            "help": "After division by the FF, and before the logarithm, the is clipped to this maximum. Enabled only if take_logarithm=1",
            "validator": float_validator,
            "type": "advanced",
        },
        "sino_normalization": {
            "default": "",
            "help": "Sinogram normalization method. Available methods are: chebyshev, subtraction, division, none. Default is none (no normalization)",
            "validator": sino_normalization_validator,
            "type": "advanced",
        },
        "sino_normalization_file": {
            "default": "",
            "help": "Path to the file when sino_normalization is either 'subtraction' or 'division'. To specify the path within a HDF5 file, the syntax is /path/to/file?path=/entry/data",
            "validator": no_validator,
            "type": "advanced",
        },
        "processes_file": {
            "default": "",
            "help": "Path to the file where some operations should be stored for later use. By default it is 'xxx_nabu_processes.h5'",
            "validator": optional_output_file_path_validator,
            "type": "advanced",
        },
        "sino_rings_correction": {
            "default": "",
            "help": "Sinogram rings removal method. Default (empty) is None. Available are: None, munch. See also: sino_rings_options",
            "validator": sino_deringer_methods,
            "type": "optional",
        },
        "sino_rings_options": {
            "default": "sigma=1.0 ; levels=10",
            "help": "Options for sinogram rings correction methods. The parameters are separated by commas and passed as 'name=value', for example: sigma=1.0;levels=10. Mind the semicolon separator (;).",
            "validator": generic_options_validator,
            "type": "advanced",
        },
        "rotate_projections": {
            "default": "",
            "help": "Whether to rotate each projection image with a certain angle (in degree). By default (empty) no rotation is done.",
            "validator": optional_nonzero_float_validator,
            "type": "advanced",
        },
        "rotate_projections_center": {
            "default": "",
            "help": "Center of rotation when 'rotate_projections' is non-empty. By default the center of rotation is the middle of each radio, i.e ((Nx-1)/2.0, (Ny-1)/2.0).",
            "validator": optional_tuple_of_floats_validator,
            "type": "advanced",
        },
        "tilt_correction": {
            "default": "",
            "help": "Detector tilt correction. Default (empty) means no tilt correction.\nThe following values can be provided for automatic tilt estimation, in this case, the projection images are rotated by the found tilt value:\n - A scalar value: tilt correction angle in degrees\n - 1d-correlation: auto-detect tilt with the 1D correlation method (fastest, but works best for small tilts)\n - fft-polar: auto-detect tilt with polar FFT method (slower, but works well on all ranges of tilts)",
            "validator": tilt_validator,
            "type": "advanced",
        },
        "autotilt_options": {
            "default": "",
            "help": "Options for methods computing automatically the detector tilt. The parameters are separated by commas and passed as 'name=value', for example: low_pass=1; high_pass=20. Mind the semicolon separator (;).",
            "validator": generic_options_validator,
            "type": "advanced",
        },
    },
    "phase": {
        "method": {
            "default": "none",
            "help": "Phase retrieval method. Available are: Paganin, CTF, None",
            "validator": phase_method_validator,
            "type": "required",
        },
        "delta_beta": {
            "default": "100.0",
            "help": "Single-distance phase retrieval related parameters\n----------------------------\ndelta/beta ratio for the Paganin/CTF method",
            "validator": float_validator,
            "type": "required",
        },
        "unsharp_coeff": {
            "default": "0",
            "help": "Unsharp mask strength. The unsharped image is equal to\n  UnsharpedImage =  (1 + coeff)*originalPaganinImage - coeff * ConvolvedImage. Setting this coefficient to zero means that no unsharp mask will be applied.",
            "validator": float_validator,
            "type": "optional",
        },
        "unsharp_sigma": {
            "default": "0",
            "help": "Standard deviation of the Gaussian filter when applying an unsharp mask\nafter the phase filtering. Disabled if set to 0.",
            "validator": float_validator,
            "type": "optional",
        },
        "unsharp_method": {
            "default": "gaussian",
            "help": "Which type of unsharp mask filter to use. Available values are gaussian, laplacian and imagej. Default is gaussian.",
            "validator": unsharp_method_validator,
            "type": "optional",
        },
        "padding_type": {
            "default": "edge",
            "help": "Padding type for the filtering step in Paganin/CTF. Available are: mirror, edge, zeros",
            "validator": padding_mode_validator,
            "type": "advanced",
        },
        "ctf_geometry": {
            "default": "z1_v=None; z1_h=None; detec_pixel_size=None; magnification=True",
            "help": "Geometric parameters for CTF phase retrieval. Length units are in meters.",
            "validator": optional_string_validator,
            "type": "optional",
        },
        "ctf_advanced_params": {
            "default": "length_scale=1e-5; lim1=1e-5; lim2=0.2; normalize_by_mean=True",
            "help": "Advanced parameters for CTF phase retrieval.",
            "validator": optional_string_validator,
            "type": "advanced",
        },
    },
    "reconstruction": {
        "method": {
            "default": "FBP",
            "help": "Reconstruction method. Possible values: FBP, cone, none. If value is 'none', no reconstruction will be done.",
            "validator": reconstruction_method_validator,
            "type": "required",
        },
        "angles_file": {
            "default": "",
            "help": "In the case you want to override the angles found in the files metadata. The angles are in degree.",
            "validator": optional_file_location_validator,
            "type": "optional",
        },
        "rotation_axis_position": {
            "default": "sliding-window",
            "help": "Rotation axis position. It can be a number or the name of an estimation method (empty value means the middle of the detector).\nThe following methods are available to find automatically the Center of Rotation (CoR):\n - centered : a fast and simple auto-CoR method. It only works when the CoR is not far from the middle of the detector. It does not work for half-tomography.\n - global : a slow but robust auto-CoR.\n - sliding-window : semi-automatically find the CoR with a sliding window. You have to specify on which side the CoR is (left, center, right). Please see the 'cor_options' parameter.\n - growing-window : automatically find the CoR with a sliding-and-growing window. You can tune the option with the parameter 'cor_options'.\n - sino-coarse-to-fine: Estimate CoR from sinogram. Only works for 360 degrees scans.\n - composite-coarse-to-fine: Estimate CoR from composite multi-angle images. Only works for 360 degrees scans.",
            "validator": cor_validator,
            "type": "required",
        },
        "cor_options": {
            "default": "",
            "help": "Options for methods finding automatically the rotation axis position. The parameters are separated by commas and passed as 'name=value', for example: low_pass=1; high_pass=20. Mind the semicolon separator (;).",
            "validator": generic_options_validator,
            "type": "advanced",
        },
        "cor_slice": {
            "default": "",
            "help": "Which slice to use for estimating the Center of Rotation (CoR). This parameter can be an integer or 'top', 'middle', 'bottom'.\nIf provided, the CoR will be estimated from the correspondig sinogram, and 'cor_options' can contain the parameter 'subsampling'.",
            "validator": cor_slice_validator,
            "type": "advanced",
        },
        "axis_correction_file": {
            "default": "",
            "help": "In the case where the axis position is specified for each angle",
            "validator": optional_values_file_validator,
            "type": "advanced",
        },
        "translation_movements_file": {
            "default": "",
            "help": "A file where each line describes the horizontal and vertical translations of the sample (or detector). The order is 'horizontal, vertical'.",
            "validator": optional_values_file_validator,
            "type": "advanced",
        },
        "angle_offset": {
            "default": "0",
            "help": "Use this if you want to obtain a rotated reconstructed slice. The angle is in degrees.",
            "validator": float_validator,
            "type": "advanced",
        },
        "fbp_filter_type": {
            "default": "ramlak",
            "help": "Filter type for FBP method. Available are: none, ramlak, shepp-logan, cosine, hamming, hann, tukey, lanczos, hilbert",
            "validator": fbp_filter_name_validator,
            "type": "advanced",
        },
        "fbp_filter_cutoff": {
            "default": "1.",
            "help": "Cut-off frequency for Fourier filter used in FBP, in normalized units. Default is the Nyquist frequency 1.0",
            "validator": float_validator,
            "type": "advanced",
        },
        "source_sample_dist": {
            "default": "",
            "help": "In cone-beam geometry, distance (in meters) between the X-ray source and the center of the sample. Default is infinity.",
            "validator": optional_float_validator,
            "type": "advanced",
        },
        "sample_detector_dist": {
            "default": "",
            "help": "In cone-beam geometry, distance (in meters) between the center of the sample and the detector. Default is read from the input dataset.",
            "validator": optional_float_validator,
            "type": "advanced",
        },
        "padding_type": {
            "default": "edges",
            "help": "Padding type for FBP. Available are: zeros, edges",
            "validator": padding_mode_validator,
            "type": "optional",  # put "advanced" with default value "edges" ?
        },
        "enable_halftomo": {
            "default": "auto",
            "help": "Whether to enable half-acquisition. Default is auto. You can enable/disable it manually by setting 1 or 0.",
            "validator": boolean_or_auto_validator,
            "type": "optional",
        },
        "clip_outer_circle": {
            "default": "0",
            "help": "Whether to set to zero voxels falling outside of the reconstruction region",
            "validator": boolean_validator,
            "type": "optional",
        },
        "centered_axis": {
            "default": "0",
            "help": "If set to true, the reconstructed region is centered on the rotation axis, i.e the center of the image will be the rotation axis position.",
            "validator": boolean_validator,
            "type": "optional",
        },
        "start_x": {
            "default": "0",
            "help": "\nParameters for sub-volume reconstruction. Indices start at 0, and upper bounds are INCLUDED!\n----------------------------------------------------------------\n(x, y) are the dimension of a slice, and (z) is the 'vertical' axis\nBy default, all the volume is reconstructed slice by slice, along the axis 'z'.",
            "validator": nonnegative_integer_validator,
            "type": "optional",
        },
        "end_x": {
            "default": "-1",
            "help": "",
            "validator": integer_validator,
            "type": "optional",
        },
        "start_y": {
            "default": "0",
            "help": "",
            "validator": nonnegative_integer_validator,
            "type": "optional",
        },
        "end_y": {
            "default": "-1",
            "help": "",
            "validator": integer_validator,
            "type": "optional",
        },
        "start_z": {
            "default": "0",
            "help": "",
            "validator": slice_num_validator,
            "type": "optional",
        },
        "end_z": {
            "default": "-1",
            "help": "",
            "validator": slice_num_validator,
            "type": "optional",
        },
        "iterations": {
            "default": "200",
            "help": "\nParameters for iterative algorithms\n------------------------------------\nNumber of iterations",
            "validator": nonnegative_integer_validator,
            "type": "unsupported",
        },
        "optim_algorithm": {
            "default": "chambolle-pock",
            "help": "Optimization algorithm for iterative methods",
            "validator": optimization_algorithm_name_validator,
            "type": "unsupported",
        },
        "weight_tv": {
            "default": "1.0e-2",
            "help": "Total Variation regularization parameter for iterative methods",
            "validator": float_validator,
            "type": "unsupported",
        },
        "preconditioning_filter": {
            "default": "1",
            "help": "Whether to enable 'filter preconditioning' for iterative methods",
            "validator": boolean_validator,
            "type": "unsupported",
        },
        "positivity_constraint": {
            "default": "1",
            "help": "Whether to enforce a positivity constraint in the reconstruction.",
            "validator": boolean_validator,
            "type": "unsupported",
        },
    },
    "output": {
        "location": {
            "default": "",
            "help": "Directory where the output reconstruction is stored.",
            "validator": optional_directory_location_validator,
            "type": "required",
        },
        "file_prefix": {
            "default": "",
            "help": "File prefix. Optional, by default it is inferred from the scanned dataset.",
            "validator": optional_file_name_validator,
            "type": "optional",
        },
        "file_format": {
            "default": "hdf5",
            "help": "Output file format. Available are: hdf5, tiff, jp2, edf, vol",
            "validator": output_file_format_validator,
            "type": "optional",
        },
        "overwrite_results": {
            "default": "1",
            "help": "What to do in the case where the output file exists.\nBy default, the output data is never overwritten and the process is interrupted if the file already exists.\nSet this option to 1 if you want to overwrite the output files.",
            "validator": boolean_validator,
            "type": "required",
        },
        "tiff_single_file": {
            "default": "0",
            "help": "Whether to create a single large tiff file for the reconstructed volume.",
            "validator": boolean_validator,
            "type": "advanced",
        },
        "jpeg2000_compression_ratio": {
            "default": "",
            "help": "Compression ratio for Jpeg2000 output.",
            "validator": optional_positive_integer_validator,
            "type": "advanced",
        },
        "float_clip_values": {
            "default": "",
            "help": "Lower and upper bounds to use when converting from float32 to int. Floating point values are clipped to these (min, max) values before being cast to integer.",
            "validator": optional_tuple_of_floats_validator,
            "type": "advanced",
        },
    },
    "postproc": {
        "output_histogram": {
            "default": "0",
            "help": "Whether to compute a histogram of the volume.",
            "validator": boolean_validator,
            "type": "optional",
        },
        "histogram_bins": {
            "default": "1000000",
            "help": "Number of bins for the output histogram. Default is one million. ",
            "validator": nonnegative_integer_validator,
            "type": "advanced",
        },
    },
    "resources": {
        "method": {
            "default": "local",
            "help": "Computations distribution method. It can be:\n  - local:  run the computations on the local machine\n  - slurm: run the computations through SLURM\n  - preview: reconstruct the slices/volume as quickly as possible, possibly doing some binning.",
            "validator": distribution_method_validator,
            "type": "required",
        },
        "gpus": {
            "default": "1",
            "help": "Number of GPUs to use.",
            "validator": nonnegative_integer_validator,
            "type": "advanced",
        },
        "gpu_id": {
            "default": "",
            "help": "For method = local only. List of GPU IDs to use. This parameter overwrites 'gpus'.\nIf left blank, exactly one GPU will be used, and the best one will be picked.",
            "validator": list_of_int_validator,
            "type": "advanced",
        },
        "cpu_workers": {
            "default": "0",
            "help": "Number of 'CPU workers' for each GPU worker. It is discouraged to set this number to more than one. A value of -1 means exactly one CPU worker.",
            "validator": integer_validator,
            "type": "unsupported",
        },
        "memory_per_node": {
            "default": "90%",
            "help": "RAM memory per computing node, either in GB or in percent of the AVAILABLE (!= total) node memory.\nIf several workers share the same node, their combined memory usage will not exceed this number.",
            "validator": resources_validator,
            "type": "unsupported",
        },
        "threads_per_node": {
            "default": "100%",
            "help": "Number of threads to allocate on each node, either a number or a percentage of the available threads",
            "validator": resources_validator,
            "type": "unsupported",
        },
        "queue": {
            "default": "gpu",
            "help": "\nParameters exclusive to the 'slurm' distribution method\n------------------------------------------------------\nName of the SLURM partition ('queue'). Full list is obtained with 'scontrol show partition'",
            "validator": nonempty_string_validator,
            "type": "unsupported",
        },
        "walltime": {
            "default": "01:00:00",
            "help": "Time limit for the SLURM resource allocation, in the format Hours:Minutes:Seconds",
            "validator": walltime_validator,
            "type": "unsupported",
        },
    },
    "pipeline": {
        "save_steps": {
            "default": "",
            "help": "Save intermediate results. This is a list of comma-separated processing steps, for ex: flatfield, phase, sinogram.\nEach step generates a HDF5 file in the form name_file_prefix.hdf5 (ex. 'sinogram_file_prefix.hdf5')",
            "validator": optional_string_validator,
            "type": "optional",
        },
        "resume_from_step": {
            "default": "",
            "help": "Resume the processing from a previously saved processing step. The corresponding file must exist in the output directory.",
            "validator": optional_string_validator,
            "type": "optional",
        },
        "steps_file": {
            "default": "",
            "help": "File where the intermediate processing steps are written. By default it is empty, and intermediate processing steps are written in the same directory as the reconstructions, with a file prefix, ex. sinogram_mydataset.hdf5.",
            "validator": optional_output_file_path_validator,
            "type": "advanced",
        },
        "verbosity": {
            "default": "2",
            "help": "Level of verbosity of the processing. 0 = terse, 3 = much information.",
            "validator": logging_validator,
            "type": "optional",
        },
    },
    # This section will be removed in the future (for now it is deprecated)
    "about": {},
}

renamed_keys = {
    "marge": {
        "section": "phase",
        "new_name": "margin",
        "since": "2020.2.0",
        "message": "Option 'marge' has been renamed 'margin' in [phase]",
    },
    "overwrite_results": {
        "section": "about",
        "new_name": "overwrite_results",
        "new_section": "output",
        "since": "2020.3.0",
        "message": "Option 'overwrite_results' was moved from section [about] to section [output]",
    },
    "nabu_config_version": {
        "section": "about",
        "new_name": "",
        "new_section": "about",
        "since": "2020.3.1",
        "message": "Option 'nabu_config_version' was removed.",
    },
    "nabu_version": {
        "section": "about",
        "new_name": "",
        "new_section": "about",
        "since": "2021.1.0",
        "message": "Option 'nabu_config' was removed.",
    },
    "verbosity": {
        "section": "about",
        "new_name": "verbosity",
        "new_section": "pipeline",
        "since": "2021.1.0",
        "message": "Option 'verbosity' was moved from section [about] to section [pipeline]",
    },
    "flatfield_enabled": {
        "section": "preproc",
        "new_name": "flatfield",
        "since": "2021.2.0",
        "message": "Option 'flatfield_enabled' has been renamed 'flatfield' in [preproc]",
    },
}



from ...resources.logger import LoggerOrPrint
import numpy as np

logger = LoggerOrPrint(None)


def find_mirror_indexes(angles_deg, tolerance_factor=1.0):
    """return a list of indexes where the ith elememnt
    contains the index of the angles_deg array element which has the value the closest
    to angles_deg[i] + 180. It is used for padding in halftomo.

    Parameters:
    -----------

    angles_deg: a nd.array of floats

    tolerance: float
       if the mirror positions are not within a distance less than tolerance fro;
       the ideal position a warning is raised

    """
    av_step = abs(np.diff(angles_deg).mean())
    tolerance = av_step * tolerance_factor

    tmp_mirror_angles_deg = angles_deg + 180

    mirror_angle_relative_indexes = (abs(abs(np.mod(tmp_mirror_angles_deg[:, None] - angles_deg, 360) - 180))).argmax(
        axis=-1
    )

    mirror_values = angles_deg[mirror_angle_relative_indexes]
    differences = abs(np.mod(mirror_values - angles_deg, 360) - 180)

    if differences.max() > tolerance:
        logger.warning(
            f"""In function  find_mirror_indexes the mirror position are far beyon tolerance from ideal position
                            tolerance is {tolerance} given by average step {av_step} and tolerance_factor {tolerance_factor}
                            and the maximum error is {differences.max()}
        """
        )

    return mirror_angle_relative_indexes



from nabu.pipeline.helical import gridded_accumulator, span_strategy
from nabu.testutils import get_data

import pytest
import numpy as np
import os
import h5py


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    # This is a helical dataset derived
    # from "crayon" dataset, using 5 slices and covering 2.5 x 360 angular span
    # in halftomo, with vertical translations.

    helical_dataset = get_data("small_sparse_helical_dataset.npz")

    helical_dataset = dict([item for item in helical_dataset.items()])
    # the radios, in the dataset file, are stored by swapping angular and x dimension
    # so that the fast running dimension runs over the projections.
    # Due to the sparsity of the dataset, where only an handful of slices
    # has signal, this gives a much better compression even when the axis is translating
    # vertically
    helical_dataset["radios"] = np.array(np.swapaxes(helical_dataset["radios_transposed"], 0, 2))
    del helical_dataset["radios_transposed"]

    # adding members: radios,  dark, flats, z_pix_per_proj, x_pix_per_proj, projection_angles_deg, pixel_size_mm, phase_margin_pix,
    #  weigth_field=weigth_field, double_flat
    dataset_keys = [
        "dark",
        "flats",
        "z_pix_per_proj",
        "x_pix_per_proj",
        "projection_angles_deg",
        "pixel_size_mm",
        "phase_margin_pix",
        "weights_field",
        "double_flat",
        "rotation_axis_position",
        "detector_shape_vh",
        "result_inset",
        "radios",
    ]
    for key in dataset_keys:
        setattr(cls, key, helical_dataset[key])

    cls.rtol_regridded = 1.0e-6


@pytest.mark.usefixtures("bootstrap")
class TestGriddedAccumulator:
    """
    Test the GriddedAccumulator.
    Rebuilds the sinogram for some selected slices of the crayon dataset
    """

    def test_regridding(self):
        span_info = span_strategy.SpanStrategy(
            z_pix_per_proj=self.z_pix_per_proj,
            x_pix_per_proj=self.x_pix_per_proj,
            detector_shape_vh=self.detector_shape_vh,
            phase_margin_pix=self.phase_margin_pix,
            projection_angles_deg=self.projection_angles_deg,
            require_redundancy=True,
            pixel_size_mm=self.pixel_size_mm,
            logger=None,
        )

        # I would like to reconstruct from feaseable height 15 to feaseable  height 18
        # relatively to the first doable slice in the vertical translation direction
        # I get the heights in the detector frame of the first and of the last

        reconstruction_space = gridded_accumulator.get_reconstruction_space(
            span_info=span_info, min_scanwise_z=15, end_scanwise_z=18, phase_margin_pix=self.phase_margin_pix
        )

        chunk_info = span_info.get_chunk_info((reconstruction_space.my_z_min, reconstruction_space.my_z_end))

        sub_region = (
            reconstruction_space.my_z_min - self.phase_margin_pix,
            reconstruction_space.my_z_end + self.phase_margin_pix,
        )

        ## useful projections
        proj_num_start, proj_num_end = chunk_info.angle_index_span

        # the first of the chunk angular range
        my_first_pnum = proj_num_start

        self.accumulator = gridded_accumulator.GriddedAccumulator(
            gridded_radios=reconstruction_space.gridded_radios,
            gridded_weights=reconstruction_space.gridded_cumulated_weights,
            diagnostic_radios=reconstruction_space.diagnostic_radios,
            diagnostic_weights=reconstruction_space.diagnostic_weights,
            diagnostic_angles=reconstruction_space.diagnostic_proj_angle,
            dark=self.dark,
            flat_indexes=[0, 7501],
            flats=self.flats,
            weights=self.weights_field,
            double_flat=self.double_flat,
        )

        # splitting in sub ranges of 100 projections
        n_granularity = 100
        pnum_start_list = list(np.arange(proj_num_start, proj_num_end, n_granularity))
        pnum_end_list = pnum_start_list[1:] + [proj_num_end]

        for pnum_start, pnum_end in zip(pnum_start_list, pnum_end_list):
            start_in_chunk = pnum_start - my_first_pnum
            end_in_chunk = pnum_end - my_first_pnum

            self._read_data_and_apply_flats(
                slice(pnum_start, pnum_end), slice(start_in_chunk, end_in_chunk), chunk_info, sub_region, span_info
            )

        res = reconstruction_space.gridded_radios / reconstruction_space.gridded_cumulated_weights

        # check only a sub part to avoid further increasing of the file on edna site
        errmax = np.max(np.abs(res[:200, 1, -500:] - self.result_inset) / np.max(res))
        assert errmax < self.rtol_regridded, "Max error is too high"

        # uncomment this to see
        # h5py.File("processed_sinogram.h5","w")["sinogram"] = res

    def _read_data_and_apply_flats(self, sub_total_prange_slice, subchunk_slice, chunk_info, sub_region, span_info):
        my_integer_shifts_v = chunk_info.integer_shift_v[subchunk_slice]
        fract_complement_shifts_v = chunk_info.fract_complement_to_integer_shift_v[subchunk_slice]
        x_shifts_list = chunk_info.x_pix_per_proj[subchunk_slice]
        subr_start_z, subr_end_z = sub_region

        subr_start_z_list = subr_start_z - my_integer_shifts_v
        subr_end_z_list = subr_end_z - my_integer_shifts_v + 1

        dtasrc_start_z = max(0, subr_start_z_list.min())
        dtasrc_end_z = min(span_info.detector_shape_vh[0], subr_end_z_list.max())

        data_raw = self.radios[sub_total_prange_slice, slice(dtasrc_start_z, dtasrc_end_z), :]

        subsampling_file_slice = sub_total_prange_slice

        # my_subsampled_indexes = self.chunk_reader._sorted_files_indices[subsampling_file_slice]
        my_subsampled_indexes = (np.arange(10000))[subsampling_file_slice]

        self.accumulator.extract_preprocess_with_flats(
            subchunk_slice,
            my_subsampled_indexes,
            chunk_info,
            np.array((subr_start_z, subr_end_z), "i"),
            np.array((dtasrc_start_z, dtasrc_end_z), "i"),
            data_raw,
        )



import warnings
from shutil import copy as copy_file
from os import path
from h5py import VirtualSource, VirtualLayout
from tomoscan.io import HDF5File
from ..resources.logger import Logger, LoggerOrPrint
from ..io.utils import get_first_hdf5_entry
from .cli_configs import ZSplitConfig
from .utils import parse_params_values

warnings.warn(
    "This command-line utility is intended as a temporary solution. Please do not rely too much on it.", Warning
)


def _get_z_translations(fname, entry):
    z_path = path.join(entry, "sample", "z_translation")
    with HDF5File(fname, "r") as fid:
        z_transl = fid[z_path][:]
    return z_transl


class NXZSplitter:
    def __init__(self, fname, output_dir, n_stages=None, entry=None, logger=None, use_virtual_dataset=False):
        self.fname = fname
        self._ext = path.splitext(fname)[-1]
        self.output_dir = output_dir
        self.n_stages = n_stages
        if entry is None:
            entry = get_first_hdf5_entry(fname)
        self.entry = entry
        self.logger = LoggerOrPrint(logger)
        self.use_virtual_dataset = use_virtual_dataset

    def _patch_nx_file(self, fname, mask):
        orig_fname = self.fname
        detector_path = path.join(self.entry, "instrument", "detector")
        sample_path = path.join(self.entry, "sample")
        with HDF5File(fname, "a") as fid:

            def patch_nx_entry(name):
                newval = fid[name][mask]
                del fid[name]
                fid[name] = newval

            detector_entries = [
                path.join(detector_path, what) for what in ["count_time", "image_key", "image_key_control"]
            ]
            sample_entries = [
                path.join(sample_path, what)
                for what in ["rotation_angle", "x_translation", "y_translation", "z_translation"]
            ]
            for what in detector_entries + sample_entries:
                self.logger.debug("Patching %s" % what)
                patch_nx_entry(what)
            # Patch "data" using a virtual dataset
            self.logger.debug("Patching data")
            data_path = path.join(detector_path, "data")
            if self.use_virtual_dataset:
                data_shape = fid[data_path].shape
                data_dtype = fid[data_path].dtype
                new_data_shape = (int(mask.sum()),) + data_shape[1:]
                vlayout = VirtualLayout(shape=new_data_shape, dtype=data_dtype)
                vsource = VirtualSource(orig_fname, name=data_path, shape=data_shape, dtype=data_dtype)
                vlayout[:] = vsource[mask, :, :]
                del fid[data_path]
                fid[detector_path].create_virtual_dataset("data", vlayout)

        if not (self.use_virtual_dataset):
            data_path = path.join(self.entry, "instrument", "detector", "data")
            with HDF5File(orig_fname, "r") as fid:
                data_arr = fid[data_path][mask, :, :]  # Actually load data. Heavy !
            with HDF5File(fname, "a") as fid:
                del fid[data_path]
                fid[data_path] = data_arr

    def z_split(self):
        """
        Split a HDF5-NX file according to different z_translation.
        """
        z_transl = _get_z_translations(self.fname, self.entry)
        different_z = set(z_transl)
        n_z = len(different_z)
        self.logger.info("Detected %d different z values: %s" % (n_z, str(different_z)))
        if n_z <= 1:
            raise ValueError("Detected only %d z-value. Stopping." % n_z)
        if self.n_stages is not None and self.n_stages != n_z:
            raise ValueError("Expected %d different stages, but I detected %d" % (self.n_stages, n_z))
        masks = [(z_transl == z) for z in different_z]
        for i_z, mask in enumerate(masks):
            fname_curr_z = path.join(
                self.output_dir, path.splitext(path.basename(self.fname))[0] + str("_%06d" % i_z) + self._ext
            )
            self.logger.info("Creating %s" % fname_curr_z)
            copy_file(self.fname, fname_curr_z)
            self._patch_nx_file(fname_curr_z, mask)


def zsplit():
    # Parse arguments
    args = parse_params_values(
        ZSplitConfig, parser_description="Split a HDF5-Nexus file according to z translation (z-series)"
    )
    # Sanitize arguments
    fname = args["input_file"]
    output_dir = args["output_directory"]
    loglevel = args["loglevel"].upper()
    entry = args["entry"]
    if len(entry) == 0:
        entry = None
    n_stages = args["n_stages"]
    if n_stages < 0:
        n_stages = None
    use_virtual_dataset = bool(args["use_virtual_dataset"])
    # Instantiate and execute
    logger = Logger("NX_z-splitter", level=loglevel, logfile="nxzsplit.log")
    nx_splitter = NXZSplitter(
        fname, output_dir, n_stages=n_stages, entry=entry, logger=logger, use_virtual_dataset=use_virtual_dataset
    )
    nx_splitter.z_split()


if __name__ == "__main__":
    zsplit()



import numpy as np
from numpy.polynomial.polynomial import Polynomial, polyval
from .alignment import AlignmentBase, plt


class DetectorTranslationAlongBeam(AlignmentBase):
    def find_shift(
        self,
        img_stack: np.ndarray,
        img_pos: np.array,
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
        return_shifts=False,
        use_adjacent_imgs=False,
    ):
        """Find the vertical and horizontal shifts for translations of the
        detector along the beam direction.

        These shifts are in pixels-per-unit-translation, and they are due to
        the misalignment of the translation stage, with respect to the beam
        propagation direction.

        To compute the vertical and horizontal tilt angles from the obtained `shift_pix`:

        >>> tilt_deg = np.rad2deg(np.arctan(shift_pix * pixel_size))

        where `pixel_size` and and the input parameter `img_pos` have to be
        expressed in the same units.

        Parameters
        ----------
        img_stack: numpy.ndarray
            A stack of images (usually 4) at different distances
        img_pos: numpy.ndarray
            Position of the images along the translation axis
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`.
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`.
        return_shifts: boolean, optional
            Adds a third returned argument, containing the pixel shifts of each
            image with respect to the first one in the stack. Defaults to False.
        use_adjacent_imgs: boolean, optional
            Compute correlation between adjacent images.
            It can be used when dealing with large shifts, to avoid overflowing the shift.
            This option allows to replicate the behavior of the reference function `alignxc.m`
            However, it is detrimental to shift fitting accuracy. Defaults to False.

        Returns
        -------
        coeff_v: float
            Estimated vertical shift in pixel per unit-distance of the detector translation.
        coeff_h: float
            Estimated horizontal shift in pixel per unit-distance of the detector translation.
        shifts_vh: list, optional
            The pixel shifts of each image with respect to the first image in the stack.
            Activated if return_shifts is True.

        Examples
        --------
        The following example creates a stack of shifted images, and retrieves the computed shift.
        Here we use a high-pass filter, due to the presence of some low-frequency noise component.

        >>> import numpy as np
        ... import scipy as sp
        ... import scipy.ndimage
        ... from nabu.preproc.alignment import  DetectorTranslationAlongBeam
        ...
        ... tr_calc = DetectorTranslationAlongBeam()
        ...
        ... stack = np.zeros([4, 512, 512])
        ...
        ... # Add low frequency spurious component
        ... for i in range(4):
        ...     stack[i, 200 - i * 10, 200 - i * 10] = 1
        ... stack = sp.ndimage.filters.gaussian_filter(stack, [0, 10, 10.0]) * 100
        ...
        ... # Add the feature
        ... x, y = np.meshgrid(np.arange(stack.shape[-1]), np.arange(stack.shape[-2]))
        ... for i in range(4):
        ...     xc = x - (250 + i * 1.234)
        ...     yc = y - (250 + i * 1.234 * 2)
        ...     stack[i] += np.exp(-(xc * xc + yc * yc) * 0.5)
        ...
        ... # Image translation along the beam
        ... img_pos = np.arange(4)
        ...
        ... # Find the shifts from the features
        ... shifts_v, shifts_h = tr_calc.find_shift(stack, img_pos, high_pass=1.0)
        ... print(shifts_v, shifts_h)
        >>> ( -2.47 , -1.236 )

        and the following commands convert the shifts in angular tilts:

        >>> tilt_v_deg = np.rad2deg(np.arctan(shifts_v * pixel_size))
        >>> tilt_h_deg = np.rad2deg(np.arctan(shifts_h * pixel_size))

        To enable the legacy behavior of `alignxc.m` (correlation between adjacent images):

        >>> shifts_v, shifts_h = tr_calc.find_shift(stack, img_pos, use_adjacent_imgs=True)

        To plot the correlation shifts and the fitted straight lines for both directions:

        >>> tr_calc = DetectorTranslationAlongBeam(verbose=True)
        ... shifts_v, shifts_h = tr_calc.find_shift(stack, img_pos)
        """
        self._check_img_stack_size(img_stack, img_pos)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        num_imgs = img_stack.shape[0]
        img_shape = img_stack.shape[-2:]
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_stack = self._prepare_image(img_stack, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        # do correlations
        ccs = [
            self._compute_correlation_fft(
                img_stack[ii - 1 if use_adjacent_imgs else 0, ...],
                img_stack[ii, ...],
                padding_mode,
                high_pass=high_pass,
                low_pass=low_pass,
            )
            for ii in range(1, num_imgs)
        ]

        img_shape = img_stack.shape[-2:]
        cc_vs = np.fft.fftfreq(img_shape[-2], 1 / img_shape[-2])
        cc_hs = np.fft.fftfreq(img_shape[-1], 1 / img_shape[-1])

        shifts_vh = np.zeros((num_imgs, 2))
        for ii, cc in enumerate(ccs):
            (f_vals, fv, fh) = self.extract_peak_region_2d(cc, peak_radius=peak_fit_radius, cc_vs=cc_vs, cc_hs=cc_hs)
            shifts_vh[ii + 1, :] = self.refine_max_position_2d(f_vals, fv, fh)

        if use_adjacent_imgs:
            shifts_vh = np.cumsum(shifts_vh, axis=0)

        # Polynomial.fit is supposed to be more numerically stable than polyfit
        # (according to numpy)
        coeffs_v = Polynomial.fit(img_pos, shifts_vh[:, 0], deg=1).convert().coef
        coeffs_h = Polynomial.fit(img_pos, shifts_vh[:, 1], deg=1).convert().coef

        if self.verbose:
            self.logger.info(
                "Fitted pixel shifts per unit-length: vertical = %f, horizontal = %f" % (coeffs_v[1], coeffs_h[1])
            )
            f, axs = plt.subplots(1, 2)
            self._add_plot_window(f, ax=axs)
            axs[0].scatter(img_pos, shifts_vh[:, 0])
            axs[0].plot(img_pos, polyval(img_pos, coeffs_v), "-C1")
            axs[0].set_title("Vertical shifts")
            axs[1].scatter(img_pos, shifts_vh[:, 1])
            axs[1].plot(img_pos, polyval(img_pos, coeffs_h), "-C1")
            axs[1].set_title("Horizontal shifts")
            plt.show(block=False)

        if return_shifts:
            return coeffs_v[1], coeffs_h[1], shifts_vh
        else:
            return coeffs_v[1], coeffs_h[1]



from ...utils import *
from ...io.writer import Writers, NXProcessWriter
from ...io.tiffwriter_zmm import TIFFWriter
from ...resources.logger import LoggerOrPrint
from ...resources.utils import is_hdf5_extension
from os import path, mkdir
from ...utils import check_supported
from ..fallback_utils import WriterConfigurator
from ..params import files_formats

Writers["tif"] = TIFFWriter
Writers["tiff"] = TIFFWriter


class WriterConfiguratorHelical(WriterConfigurator):
    def _get_initial_writer_kwarg(self):
        if self.file_format in ["tif", "tiff"]:
            return {"heights_above_stage_mm": self.heights_above_stage_mm}
        else:
            return {}

    def __init__(
        self,
        output_dir,
        file_prefix,
        file_format="hdf5",
        overwrite=False,
        start_index=None,
        logger=None,
        nx_info=None,
        write_histogram=False,
        histogram_entry="entry",
        writer_options=None,
        extra_options=None,
        heights_above_stage_mm=None,
    ):
        self.heights_above_stage_mm = heights_above_stage_mm
        self.file_format = file_format
        super().__init__(
            output_dir,
            file_prefix,
            file_format=file_format,
            overwrite=overwrite,
            start_index=start_index,
            logger=logger,
            nx_info=nx_info,
            write_histogram=write_histogram,
            histogram_entry=histogram_entry,
            writer_options=writer_options,
            extra_options=extra_options,
        )



flatfield_modes = {
    "true": True,
    "1": True,
    "false": False,
    "0": False,
    "forced": "force-load",
    "force-load": "force-load",
    "force-compute": "force-compute",
}

phase_retrieval_methods = {
    "": None,
    "none": None,
    "paganin": "paganin",
    "tie": "paganin",
    "ctf": "CTF",
}

unsharp_methods = {
    "gaussian": "gaussian",
    "log": "log",
    "laplacian": "log",
    "imagej": "imagej",
    "none": None,
    "": None,
}

padding_modes = {
    "edges": "edge",
    "edge": "edge",
    "mirror": "mirror",
    "zeros": "zeros",
    "zero": "zeros",
}

reconstruction_methods = {"fbp": "FBP", "cone": "cone", "conic": "cone", "none": None, "": None}

fbp_filters = {
    "ramlak": "ramlak",
    "ram-lak": "ramlak",
    "none": None,
    "": None,
    "shepp-logan": "shepp-logan",
    "cosine": "cosine",
    "hamming": "hamming",
    "hann": "hann",
    "tukey": "tukey",
    "lanczos": "lanczos",
    "hilbert": "hilbert",
}

iterative_methods = {
    "tv": "TV",
    "wavelets": "wavelets",
    "l2": "L2",
    "ls": "L2",
    "sirt": "SIRT",
    "em": "EM",
}

optim_algorithms = {
    "chambolle": "chambolle-pock",
    "chambollepock": "chambolle-pock",
    "fista": "fista",
}

files_formats = {
    "h5": "hdf5",
    "hdf5": "hdf5",
    "nexus": "hdf5",
    "nx": "hdf5",
    "npy": "npy",
    "npz": "npz",
    "tif": "tiff",
    "tiff": "tiff",
    "jp2": "jp2",
    "jp2k": "jp2",
    "j2k": "jp2",
    "jpeg2000": "jp2",
    "edf": "edf",
    "vol": "vol",
}

distribution_methods = {
    "local": "local",
    "slurm": "slurm",
    "": "local",
    "preview": "preview",
}

log_levels = {
    "0": "error",
    "1": "warning",
    "2": "info",
    "3": "debug",
}

sino_normalizations = {
    "none": None,
    "": None,
    "chebyshev": "chebyshev",
    "subtraction": "subtraction",
    "division": "division",
}

cor_methods = {
    "auto": "centered",
    "centered": "centered",
    "global": "global",
    "sliding-window": "sliding-window",
    "sliding window": "sliding-window",
    "growing-window": "growing-window",
    "growing window": "growing-window",
    "sino-coarse-to-fine": "sino-coarse-to-fine",
    "composite-coarse-to-fine": "composite-coarse-to-fine",
    "near": "composite-coarse-to-fine",
}

tilt_methods = {
    "1d-correlation": "1d-correlation",
    "1dcorrelation": "1d-correlation",
    "polarfft": "fft-polar",
    "polar-fft": "fft-polar",
    "fft-polar": "fft-polar",
}

rings_methods = {
    "none": None,
    "": None,
    "munch": "munch",
}

detector_distortion_correction_methods = {"none": None, "": None, "identity": "identity", "map_xz": "map_xz"}


radios_rotation_mode = {
    "none": None,
    "": None,
    "chunk": "chunk",
    "chunks": "chunk",
    "full": "full",
}



from os import path
from tempfile import mkdtemp
from shutil import rmtree
import pytest
import numpy as np
from silx.io import get_data
from tomoscan.esrf.scan.hdf5scan import ImageKey
from nabu.testutils import generate_nx_dataset
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer


test_nxflatfield_scenarios = [
    {
        "name": "simple",
        "flats_pos": [slice(1, 6)],
        "darks_pos": [slice(0, 1)],
        "output_dir": None,
    },
    {
        "name": "simple_with_save",
        "flats_pos": [slice(1, 6)],
        "darks_pos": [slice(0, 1)],
        "output_dir": None,
    },
    {
        "name": "multiple_with_save",
        "flats_pos": [slice(0, 10), slice(30, 40)],
        "darks_pos": [slice(95, 100)],
        "output_dir": path.join("{tempdir}", "output_reduced"),
    },
]


# parametrize with fixture and "params=" will launch a new class for each scenario.
# the attributes set to "cls" will remain for all the tests done in this class
# with the current scenario.
@pytest.fixture(scope="class", params=test_nxflatfield_scenarios)
def bootstrap(request):
    cls = request.cls
    cls.n_projs = 265
    cls.params = request.param
    cls.tempdir = mkdtemp(prefix="nabu_")
    yield
    rmtree(cls.tempdir)


@pytest.mark.usefixtures("bootstrap")
class TestNXFlatField:
    _reduction_func = {"flats": np.median, "darks": np.mean}

    def get_nx_filename(self):
        return path.join(self.tempdir, "dataset_" + self.params["name"] + ".nx")

    def get_image_key(self):
        keys = np.zeros(self.n_projs, np.int32)
        for what, val in [("flats_pos", ImageKey.FLAT_FIELD.value), ("darks_pos", ImageKey.DARK_FIELD.value)]:
            for pos in self.params[what]:
                keys[pos.start : pos.stop] = val
        return keys

    @staticmethod
    def check_image_key(dataset_info, frame_type, expected_slices):
        data_slices = dataset_info.get_data_slices(frame_type)
        assert set(map(str, data_slices)) == set(map(str, expected_slices))

    def test_nxflatfield(self):
        dataset_fname = self.get_nx_filename()
        image_key = self.get_image_key()
        generate_nx_dataset(dataset_fname, image_key)

        dataset_info = HDF5DatasetAnalyzer(dataset_fname)

        # When parsing a "raw" dataset, flats/darks are a series of images.
        # dataset_info.flats is a dictionary where each key is the index of the frame.
        # For example dataset_info.flats.keys() = [10, 11, 12, 13, ..., 19, 1200, 1201, 1202, ..., 1219]
        for frame_type in ["darks", "flats"]:
            self.check_image_key(dataset_info, frame_type, self.params[frame_type + "_pos"])

        output_dir = self.params.get("output_dir", None)
        if output_dir is not None:
            output_dir = output_dir.format(tempdir=self.tempdir)
        update_dataset_info_flats_darks(dataset_info, True, output_dir=output_dir)
        # After reduction (median/mean), the flats/darks are located in another file.
        # median(series_1) goes to entry/flats/idx1, mean(series_2) goes to entry/flats/idx2, etc.
        assert set(dataset_info.flats.keys()) == set(s.start for s in self.params["flats_pos"])
        assert set(dataset_info.darks.keys()) == set(s.start for s in self.params["darks_pos"])

        # Check that the computations were correct
        # Loads the entire volume in memory ! So keep the data volume small for the tests
        data_volume = get_data(dataset_info.dataset_hdf5_url)
        expected_flats = {}
        for s in self.params["flats_pos"]:
            expected_flats[s.start] = self._reduction_func["flats"](data_volume[s.start : s.stop], axis=0)
        expected_darks = {}
        for s in self.params["darks_pos"]:
            expected_darks[s.start] = self._reduction_func["darks"](data_volume[s.start : s.stop], axis=0)

        flats = {idx: get_data(dataset_info.flats[idx]) for idx in dataset_info.flats.keys()}
        for idx in flats.keys():
            assert np.allclose(flats[idx], expected_flats[idx])
        darks = {idx: get_data(dataset_info.darks[idx]) for idx in dataset_info.darks.keys()}
        for idx in darks.keys():
            assert np.allclose(darks[idx], expected_darks[idx])
